import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

# Graph-based Knowledge Tracing: Modeling Student Proficiency Using Graph Neural Network.
# For more information, please refer to https://dl.acm.org/doi/10.1145/3350546.3352513
# Author: jhljx
# Email: jhljx8918@gmail.com


# Multi-Layer Perceptron(MLP) layer
class MLP(nn.Module):
    """Two-layer fully-connected ReLU net with batch norm."""

    def __init__(self, input_dim, hidden_dim, output_dim, dropout=0., bias=True):
        super(MLP, self).__init__()
        self.fc1 = nn.Linear(input_dim, hidden_dim, bias=bias)
        self.fc2 = nn.Linear(hidden_dim, output_dim, bias=bias)
        self.norm = nn.BatchNorm1d(output_dim)
        # the paper said they added Batch Normalization for the output of MLPs, as shown in Section 4.2
        self.dropout = dropout
        self.output_dim = output_dim
        self.init_weights()

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_normal_(m.weight.data)
                m.bias.data.fill_(0.1)
            elif isinstance(m, nn.BatchNorm1d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def batch_norm(self, inputs):
        if inputs.numel() == self.output_dim or inputs.numel() == 0:
            # batch_size == 1 or 0 will cause BatchNorm error, so return the input directly
            return inputs
        if len(inputs.size()) == 3:
            x = inputs.view(inputs.size(0) * inputs.size(1), -1)
            x = self.norm(x)
            return x.view(inputs.size(0), inputs.size(1), -1)
        else:  # len(input_size()) == 2
            return self.norm(inputs)

    def forward(self, inputs):
        x = F.relu(self.fc1(inputs))
        x = F.dropout(x, self.dropout, training=self.training)  # pay attention to add training=self.training
        x = F.relu(self.fc2(x))
        return self.batch_norm(x)


class EraseAddGate(nn.Module):
    """
    Erase & Add Gate module
    NOTE: this erase & add gate is a bit different from that in DKVMN.
    For more information about Erase & Add gate, please refer to the paper "Dynamic Key-Value Memory Networks for Knowledge Tracing"
    The paper can be found in https://arxiv.org/abs/1611.08108
    """

    def __init__(self, feature_dim, concept_num, bias=True):
        super(EraseAddGate, self).__init__()
        # weight
        self.weight = nn.Parameter(torch.rand(concept_num))
        self.reset_parameters()
        # erase gate
        self.erase = nn.Linear(feature_dim, feature_dim, bias=bias)
        # add gate
        self.add = nn.Linear(feature_dim, feature_dim, bias=bias)

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(0))
        self.weight.data.uniform_(-stdv, stdv)

    def forward(self, x):
        r"""
        Params:
            x: input feature matrix
        Shape:
            x: [batch_size, concept_num, feature_dim]
            res: [batch_size, concept_num, feature_dim]
        Return:
            res: returned feature matrix with old information erased and new information added
        The GKT paper didn't provide detailed explanation about this erase-add gate. As the erase-add gate in the GKT only has one input parameter,
        this gate is different with that of the DKVMN. We used the input matrix to build the erase and add gates, rather than $\mathbf{v}_{t}$ vector in the DKVMN.
        """
        erase_gate = torch.sigmoid(self.erase(x))  # [batch_size, concept_num, feature_dim]
        # self.weight.unsqueeze(dim=1) shape: [concept_num, 1]
        tmp_x = x - self.weight.unsqueeze(dim=1) * erase_gate * x
        add_feat = torch.tanh(self.add(x))  # [batch_size, concept_num, feature_dim]
        res = tmp_x + self.weight.unsqueeze(dim=1) * add_feat
        return res


class ScaledDotProductAttention(nn.Module):
    """
    Scaled Dot-Product Attention
    NOTE: Stole and modify from https://github.com/jadore801120/attention-is-all-you-need-pytorch/blob/master/transformer/Modules.py
    """

    def __init__(self, temperature, attn_dropout=0.):
        super().__init__()
        self.temperature = temperature
        self.dropout = attn_dropout

    def forward(self, q, k, mask=None):
        r"""
        Parameters:
            q: multi-head query matrix
            k: multi-head key matrix
            mask: mask matrix
        Shape:
            q: [n_head, mask_num, embedding_dim]
            k: [n_head, concept_num, embedding_dim]
        Return: attention score of all queries
        """
        attn = torch.matmul(q / self.temperature, k.transpose(1, 2))  # [n_head, mask_number, concept_num]
        if mask is not None:
            attn = attn.masked_fill(mask == 0, -1e9)
        # pay attention to add training=self.training!
        attn = F.dropout(F.softmax(attn, dim=0), self.dropout, training=self.training)  # pay attention that dim=-1 is not as good as dim=0!
        return attn


class MLPEncoder(nn.Module):
    """
    MLP encoder module.
    NOTE: Stole and modify the code from https://github.com/ethanfetaya/NRI/blob/master/modules.py
    """
    def __init__(self, input_dim, hidden_dim, output_dim, factor=True, dropout=0., bias=True):
        super(MLPEncoder, self).__init__()
        self.factor = factor
        self.mlp = MLP(input_dim * 2, hidden_dim, hidden_dim, dropout=dropout, bias=bias)
        self.mlp2 = MLP(hidden_dim, hidden_dim, hidden_dim, dropout=dropout, bias=bias)
        if self.factor:
            self.mlp3 = MLP(hidden_dim * 3, hidden_dim, hidden_dim, dropout=dropout, bias=bias)
        else:
            self.mlp3 = MLP(hidden_dim * 2, hidden_dim, hidden_dim, dropout=dropout, bias=bias)
        self.fc_out = nn.Linear(hidden_dim, output_dim)
        self.init_weights()

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_normal_(m.weight.data)
                m.bias.data.fill_(0.1)

    def node2edge(self, x, sp_send, sp_rec):
        # NOTE: Assumes that we have the same graph across all samples.
        receivers = torch.matmul(sp_rec, x)
        senders = torch.matmul(sp_send, x)
        edges = torch.cat([senders, receivers], dim=1)
        return edges

    def edge2node(self, x, sp_send_t, sp_rec_t):
        # NOTE: Assumes that we have the same graph across all samples.
        incoming = torch.matmul(sp_rec_t, x)
        return incoming

    def forward(self, inputs, sp_send, sp_rec, sp_send_t, sp_rec_t):
        r"""
        Parameters:
            inputs: input concept embedding matrix
            sp_send: one-hot encoded send-node index(sparse tensor)
            sp_rec: one-hot encoded receive-node index(sparse tensor)
            sp_send_t: one-hot encoded send-node index(sparse tensor, transpose)
            sp_rec_t: one-hot encoded receive-node index(sparse tensor, transpose)
        Shape:
            inputs: [concept_num, embedding_dim]
            sp_send: [edge_num, concept_num]
            sp_rec: [edge_num, concept_num]
            sp_send_t: [concept_num, edge_num]
            sp_rec_t: [concept_num, edge_num]
        Return:
            output: [edge_num, edge_type_num]
        """
        x = self.node2edge(inputs, sp_send, sp_rec)  # [edge_num, 2 * embedding_dim]
        x = self.mlp(x)  # [edge_num, hidden_num]
        x_skip = x

        if self.factor:
            x = self.edge2node(x, sp_send_t, sp_rec_t)  # [concept_num, hidden_num]
            x = self.mlp2(x)  # [concept_num, hidden_num]
            x = self.node2edge(x, sp_send, sp_rec)  # [edge_num, 2 * hidden_num]
            x = torch.cat((x, x_skip), dim=1)  # Skip connection  shape: [edge_num, 3 * hidden_num]
            x = self.mlp3(x)  # [edge_num, hidden_num]
        else:
            x = self.mlp2(x)  # [edge_num, hidden_num]
            x = torch.cat((x, x_skip), dim=1)  # Skip connection  shape: [edge_num, 2 * hidden_num]
            x = self.mlp3(x)  # [edge_num, hidden_num]
        output = self.fc_out(x)  # [edge_num, output_dim]
        return output


class MLPDecoder(nn.Module):
    """
    MLP decoder module.
    NOTE: Stole and modify the code from https://github.com/ethanfetaya/NRI/blob/master/modules.py
    """

    def __init__(self, input_dim, msg_hidden_dim, msg_output_dim, hidden_dim, edge_type_num, dropout=0., bias=True):
        super(MLPDecoder, self).__init__()
        self.msg_out_dim = msg_output_dim
        self.edge_type_num = edge_type_num
        self.dropout = dropout

        self.msg_fc1 = nn.ModuleList([nn.Linear(2 * input_dim, msg_hidden_dim, bias=bias) for _ in range(edge_type_num)])
        self.msg_fc2 = nn.ModuleList([nn.Linear(msg_hidden_dim, msg_output_dim, bias=bias) for _ in range(edge_type_num)])
        self.out_fc1 = nn.Linear(msg_output_dim, hidden_dim, bias=bias)
        self.out_fc2 = nn.Linear(hidden_dim, hidden_dim, bias=bias)
        self.out_fc3 = nn.Linear(hidden_dim, input_dim, bias=bias)

    def node2edge(self, x, sp_send, sp_rec):
        receivers = torch.matmul(sp_rec, x)  # [edge_num, embedding_dim]
        senders = torch.matmul(sp_send, x)  # [edge_num, embedding_dim]
        edges = torch.cat([senders, receivers], dim=-1)  # [edge_num, 2 * embedding_dim]
        return edges

    def edge2node(self, x, sp_send_t, sp_rec_t):
        # NOTE: Assumes that we have the same graph across all samples.
        incoming = torch.matmul(sp_rec_t, x)
        return incoming

    def forward(self, inputs, rel_type, sp_send, sp_rec, sp_send_t, sp_rec_t):
        r"""
        Parameters:
            inputs: input concept embedding matrix
            rel_type: inferred edge weights for all edge types from MLPEncoder
            sp_send: one-hot encoded send-node index(sparse tensor)
            sp_rec: one-hot encoded receive-node index(sparse tensor)
            sp_send_t: one-hot encoded send-node index(sparse tensor, transpose)
            sp_rec_t: one-hot encoded receive-node index(sparse tensor, transpose)
        Shape:
            inputs: [concept_num, embedding_dim]
            sp_send: [edge_num, concept_num]
            sp_rec: [edge_num, concept_num]
            sp_send_t: [concept_num, edge_num]
            sp_rec_t: [concept_num, edge_num]
        Return:
            output: [edge_num, edge_type_num]
        """
        # NOTE: Assumes that we have the same graph across all samples.
        # Node2edge
        pre_msg = self.node2edge(inputs, sp_send, sp_rec)
        all_msgs = Variable(torch.zeros(pre_msg.size(0), self.msg_out_dim, device=inputs.device))  # [edge_num, msg_out_dim]
        for i in range(self.edge_type_num):
            msg = F.relu(self.msg_fc1[i](pre_msg))
            msg = F.dropout(msg, self.dropout, training=self.training)
            msg = F.relu(self.msg_fc2[i](msg))
            msg = msg * rel_type[:, i:i + 1]
            all_msgs += msg

        # Aggregate all msgs to receiver
        agg_msgs = self.edge2node(all_msgs, sp_send_t, sp_rec_t)  # [concept_num, msg_out_dim]
        # Output MLP
        pred = F.dropout(F.relu(self.out_fc1(agg_msgs)), self.dropout, training=self.training)  # [concept_num, hidden_dim]
        pred = F.dropout(F.relu(self.out_fc2(pred)), self.dropout, training=self.training)  # [concept_num, hidden_dim]
        pred = self.out_fc3(pred)  # [concept_num, embedding_dim]
        return pred



from .base_evaltpl import BaseEvalTPL
from .prediction_evaltpl import PredictionEvalTPL
from .interpretability_evaltpl import InterpretabilityEvalTPL
from .fairness_evaltpl import FairnessEvalTPL
from .identifiability_evaltpl import IdentifiabilityEvalTPL




# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

from .callback import Callback
from typing import Tuple, List
import torch
import os
import shutil
import glob
from .baseLogger import BaseLogger
import numpy as np
from collections import namedtuple


class Metric(object):
    def __init__(self, name, type_):
        assert type_ in ['max', 'min']
        self.name = name
        self.type_ = type_
        self.best_epoch = 1
        self.best_log = dict()

        if self.type_ == 'max':
            self.best_value = -np.inf
        else:
            self.best_value = np.inf

    def better_than(self, value):
        return value < self.best_value if self.type_ == 'max' else value > self.best_value

    def update(self, epoch, value, log):
        self.best_epoch = epoch
        self.best_value = value
        self.best_log = log

class ModelCheckPoint(Callback):
    def __init__(self, metric_list: List[list], save_folder_path, save_best_only=True):
        super().__init__()
        self.save_folder_path = save_folder_path
        if not os.path.exists(self.save_folder_path):
            os.makedirs(self.save_folder_path)
        self.save_best_only = save_best_only
        self.metric_list = [Metric(name=metric_name, type_=metric_type) for metric_name, metric_type in metric_list]
        

    def on_epoch_end(self, epoch, logs=None, **kwargs):
        for metric in self.metric_list:
            if not metric.better_than(logs[metric.name]):
                metric.update(epoch=epoch, value=logs[metric.name], log=logs)
        
                self.logger.info(f"Update best epoch as {[epoch]} for {metric.name}!")
                filepath = os.path.join(self.save_folder_path, f"best-epoch-for-{metric.name}.pth")
                torch.save(obj=self.model.state_dict(), f=filepath)

        if not self.save_best_only:
            filepath = os.path.join(self.save_folder_path, f"{epoch:03d}.pth")
            torch.save(obj=self.model.state_dict(), f=filepath)            

    def on_train_end(self, logs=None, **kwargs):
        self.logger.info("==="*10)
        for metric in self.metric_list:
            self.logger.info(f"[For {metric.name}], the Best Epoch is: {metric.best_epoch}, the value={metric.best_value:.4f}")
            for cb in self.callback_list.callbacks:
                if isinstance(cb, BaseLogger):
                    cb.on_epoch_end(metric.best_epoch, logs=metric.best_log)
            # rename best epoch filename
            os.renames(
                os.path.join(self.save_folder_path, f"best-epoch-for-{metric.name}.pth"), 
                os.path.join(self.save_folder_path, f"best-epoch-{metric.best_epoch:03d}-for-{metric.name}.pth")
            )
        self.logger.info("===" * 10)



from edustudio.atom_op.raw2mid import BaseRaw2Mid
import pandas as pd
import numpy as np
import time

"""
    SLP Dataset: https://aic-fe.bnu.edu.cn/en/data/index.html
"""


class R2M_SLP_English(BaseRaw2Mid):
    """
        rawdata: https://aic-fe.bnu.edu.cn/en/data/index.html
    """
    def process(self):
        super().process()

        # for stu
        df_stu = pd.read_csv(f"{self.rawpath}/student.csv")
        df_stu.dropna(subset=['school_id'], inplace=True, how='any', axis=0)
        df_stu = df_stu[df_stu['school_id'] != 'n.a.']

        df_stu = df_stu.merge(
            pd.read_csv(f"{self.rawpath}/family.csv", index_col=False),
            on=['student_id'], how='inner'
        )

        df_stu = df_stu.merge(
                pd.read_csv(f"{self.rawpath}/school.csv"),
                on=['school_id'], how='inner'
        )

        df_stu.drop([
            'rate_of_higher_educated_teachers',
            "rate_of_teachers_with_master's_degree_and_above"
        ], inplace=True, axis=1)
        df_stu.rename(columns={
            'student_id': 'stu_id:token', 'gender': 'gender:token', 
            'school_id': 'sch_id:token', 'class_id': 'class_id:token', 
            'age_father': 'age_father:float', 'age_mother': 'age_mother:token',
            'edubg_father': 'edubg_father:token', 'edubg_mother':'edubg_mother:token', 
            'affiliation_father':'affiliation_father:token', 
            'affiliation_mother': 'affiliation_mother:token',
            'family_income': 'family_income:token', 'is_only_child':'is_only_child:token',
            'live_on_campus': 'live_on_campus:token', 
            'gathering_frequency_father':'gathering_frequency_father:token',
            'gathering_frequency_mother':'gathering_frequency_mother:token', 
            'family_traveling_times': "family_traveling_times:token", 
            'school_type': 'school_type:token',
            'dist_to_downtown': 'dist_to_downtown:float',
            #'rate_of_higher_educated_teachers': 'rate_of_higher_educated_teachers:float',
            #"rate_of_teachers_with_master's_degree_and_above": "rate_of_teachers_with_master's_degree_and_above:float",
        }, inplace=True)

        # for inter
        df_inter = pd.read_csv(f"{self.rawpath}/term-eng.csv", index_col=False, low_memory=False)
        df_inter = df_inter[(df_inter == 'n.a.').sum(axis=1) == 0].reset_index(drop=True)
        df_inter = df_inter[df_inter['concept'] != 'n.a.']
        df_inter['label'] = df_inter['score']/df_inter['full_score'].astype(float)
        
        df_exer = df_inter[['question_id', 'exam_id', 'subject_abbr', 'concept']]
        df_inter = df_inter[['student_id', 'question_id', 'score', 'full_score', 'time_access', 'label']]
        df_exer.drop_duplicates(subset=['question_id'], inplace=True)
        df_exer['concept'] = df_exer['concept'].apply(lambda x: x.split(";"))
        df_inter['time_access'] = df_inter['time_access'].apply(lambda x: self.convert2timestamp(x))

        df_inter.rename(columns={
            'student_id': 'stu_id:token', 'question_id': 'exer_id:token',
            'score': 'score:float', 'full_score':'full_score:float', 
            'time_access': 'start_timestamp:float', 'label':'label:float'
        }, inplace=True)

        df_exer.rename(columns={
             'question_id': 'exer_id:token',
             'exam_id': 'exam_id:token', 
             'subject_abbr': 'subject_abbr:token',
             'concept': 'cpt_seq:token_seq'
        }, inplace=True)

        df_inter['order_id:float'] = df_inter['start_timestamp:float'].astype(int)
        
        # save
        df_inter.to_csv(f"{self.midpath}/{self.dt}.inter.csv", index=False, encoding='utf-8')
        df_stu.to_csv(f"{self.midpath}/{self.dt}.stu.csv", index=False, encoding='utf-8')
        df_exer.to_csv(f"{self.midpath}/{self.dt}.exer.csv", index=False, encoding='utf-8')

    @staticmethod
    def convert2timestamp(dt):
        timeArray = time.strptime(dt, "%Y-%m-%d %H:%M:%S")
        timestamp = time.mktime(timeArray)
        return timestamp


r"""
KSCD
##########################################

Reference:
    Haiping Ma et al. "Knowledge-Sensed Cognitive Diagnosis for Intelligent Education Platforms" in CIKM 2022.

Reference Code:
    https://github.com/BIMK/Intelligent-Education/tree/main/KSCD_Code_F

"""

from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F
from ..utils.components import PosMLP, PosLinear


class KSCD(GDBaseModel):
    default_cfg = {
        'emb_dim': 20,
        'dropout_rate': 0.5,
        'interaction_type': 'kscd', # ['kscd', 'ncdm']
        'interaction_type_ncdm': {
            'dnn_units': [512, 256],
            'dropout_rate': 0.5,
            'activation': 'sigmoid',
            'disc_scale': 10
        },
        'interaction_type_kscd': {
            'dropout_rate': 0.0,
        }
    }
    def __init__(self, cfg):
        super().__init__(cfg)    

    def add_extra_data(self, **kwargs):
        self.Q_mat = kwargs['Q_mat'].to(self.device)

    def build_cfg(self):
        self.stu_count = self.datatpl_cfg['dt_info']['stu_count']
        self.exer_count = self.datatpl_cfg['dt_info']['exer_count']
        self.cpt_count = self.datatpl_cfg['dt_info']['cpt_count']
    
        self.lowdim = self.modeltpl_cfg['emb_dim']
        self.interaction_type = self.modeltpl_cfg['interaction_type']

    def build_model(self):
        self.stu_emb = nn.Embedding(self.stu_count, self.lowdim)
        self.cpt_emb = nn.Embedding(self.cpt_count, self.lowdim)
        self.exer_emb = nn.Embedding(self.exer_count, self.lowdim)

        if self.modeltpl_cfg['interaction_type'] == 'ncdm':
            # self.exer_disc_infer = nn.Linear(self.lowdim, 1)
            self.layer1 = nn.Linear(self.lowdim, 1)
            self.pd_net = PosMLP(
                input_dim=self.cpt_count, output_dim=1, 
                activation=self.modeltpl_cfg['interaction_type_ncdm']['activation'],
                dnn_units=self.modeltpl_cfg['interaction_type_ncdm']['dnn_units'], 
                dropout_rate=self.modeltpl_cfg['interaction_type_ncdm']['dropout_rate']
            )
        elif self.modeltpl_cfg['interaction_type'] == 'kscd':
            self.prednet_full1 = PosLinear(self.cpt_count + self.lowdim, self.cpt_count, bias=False)
            self.drop_1 = nn.Dropout(p=self.modeltpl_cfg['interaction_type_kscd']['dropout_rate'])
            self.prednet_full2 = PosLinear(self.cpt_count + self.lowdim, self.cpt_count, bias=False)
            self.drop_2 = nn.Dropout(p=self.modeltpl_cfg['interaction_type_kscd']['dropout_rate'])
            self.prednet_full3 = PosLinear(1 * self.cpt_count, 1)
        else:
            raise ValueError(f"unknown interaction_type: {self.modeltpl_cfg['interaction_type']}")

    def forward(self, stu_id, exer_id, **kwargs):
        
        stu_emb = self.stu_emb(stu_id)
        exer_emb = self.exer_emb(exer_id)
        exer_q_mat = self.Q_mat[exer_id]

        stu_ability = torch.mm(stu_emb, self.cpt_emb.weight.T).sigmoid()
        exer_diff = torch.mm(exer_emb, self.cpt_emb.weight.T).sigmoid()
        
        if self.modeltpl_cfg['interaction_type'] == 'ncdm':
            # exer_disc = self.exer_disc_infer(exer_emb).sigmoid() * self.modeltpl_cfg['interaction_type_ncdm']['disc_scale']
            exer_disc = torch.sigmoid(self.layer1(exer_emb)) * self.modeltpl_cfg['interaction_type_ncdm']['disc_scale']
            input_x = exer_disc * (stu_ability - exer_diff) * exer_q_mat
            y_pd = self.pd_net(input_x).sigmoid()
        elif self.modeltpl_cfg['interaction_type'] == 'kscd':
            batch_stu_vector = stu_ability.repeat(1, self.cpt_count).reshape(stu_ability.shape[0], self.cpt_count, stu_ability.shape[1])
            batch_exer_vector = exer_diff.repeat(1, self.cpt_count).reshape(exer_diff.shape[0], self.cpt_count, exer_diff.shape[1])

            kn_vector = self.cpt_emb.weight.repeat(stu_ability.shape[0], 1).reshape(stu_ability.shape[0], self.cpt_count, self.lowdim)

            # CD
            preference = torch.sigmoid(self.prednet_full1(torch.cat((batch_stu_vector, kn_vector), dim=2)))
            diff = torch.sigmoid(self.prednet_full2(torch.cat((batch_exer_vector, kn_vector), dim=2)))
            o = torch.sigmoid(self.prednet_full3(preference - diff))

            sum_out = torch.sum(o * exer_q_mat.unsqueeze(2), dim=1)
            count_of_concept = torch.sum(exer_q_mat, dim=1).unsqueeze(1)
            y_pd = sum_out / count_of_concept
        else:
            raise ValueError(f"unknown interaction_type: {self.modeltpl_cfg['interaction_type']}")

        return y_pd

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(stu_id=stu_id, exer_id=exer_id).flatten()
        loss = F.binary_cross_entropy(input=pd, target=label)
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)
    
    @torch.no_grad()
    def predict(self, stu_id, exer_id, **kwargs):
        return {
            'y_pd': self(stu_id=stu_id, exer_id=exer_id).flatten(),
        }

    def get_stu_status(self, stu_id=None):
        if stu_id is not None:
            stu_emb = self.stu_emb(stu_id)
        else:
            stu_emb = self.stu_emb.weight
        stu_ability = torch.mm(stu_emb, self.cpt_emb.weight.T).sigmoid()
        return stu_ability



from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F


class DKVMN(GDBaseModel):
    default_cfg = {
        'dim_s': 200,
        'size_m': 50,
        'drop_out': 0.2,
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    def _init_params(self):
        super()._init_params()
        nn.init.kaiming_normal_(self.Mk)
        nn.init.kaiming_normal_(self.Mv0)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']

    def build_model(self):
        self.k_emb_layer = nn.Embedding(self.n_item, self.modeltpl_cfg['dim_s'])
        self.Mk = nn.Parameter(torch.Tensor(self.modeltpl_cfg['size_m'], self.modeltpl_cfg['dim_s']))
        self.Mv0 = nn.Parameter(torch.Tensor(self.modeltpl_cfg['size_m'], self.modeltpl_cfg['dim_s']))

        self.v_emb_layer = nn.Embedding(self.n_item * 2, self.modeltpl_cfg['dim_s'])

        self.f_layer = nn.Linear(self.modeltpl_cfg['dim_s'] * 2, self.modeltpl_cfg['dim_s'])
        self.dropout_layer = nn.Dropout(self.modeltpl_cfg['drop_out'])
        self.p_layer = nn.Linear(self.modeltpl_cfg['dim_s'], 1)

        self.e_layer = nn.Linear(self.modeltpl_cfg['dim_s'], self.modeltpl_cfg['dim_s'])
        self.a_layer = nn.Linear(self.modeltpl_cfg['dim_s'], self.modeltpl_cfg['dim_s'])


    def forward(self, exer_seq, label_seq, **kwargs):
        batch_size = exer_seq.shape[0]
        x = exer_seq + self.n_item * label_seq
        k = self.k_emb_layer(exer_seq.long())
        v = self.v_emb_layer(x.long())

        Mvt = self.Mv0.unsqueeze(0).repeat(batch_size, 1, 1)

        Mv = [Mvt]

        w = torch.softmax(torch.matmul(k, self.Mk.T), dim=-1)

        # Write Process
        e = torch.sigmoid(self.e_layer(v))
        a = torch.tanh(self.a_layer(v))

        for et, at, wt in zip(
                e.permute(1, 0, 2), a.permute(1, 0, 2), w.permute(1, 0, 2)
        ):
            Mvt = Mvt * (1 - (wt.unsqueeze(-1) * et.unsqueeze(1))) + \
                  (wt.unsqueeze(-1) * at.unsqueeze(1))
            Mv.append(Mvt)

        Mv = torch.stack(Mv, dim=1)

        # Read Process
        f = torch.tanh(
            self.f_layer(
                torch.cat(
                    [
                        (w.unsqueeze(-1) * Mv[:, :-1]).sum(-2),
                        k
                    ],
                    dim=-1
                )
            )
        )
        p = self.p_layer(self.dropout_layer(f))

        y_pd = torch.sigmoid(p)

        return y_pd.squeeze(-1)

    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, 1:]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, 1:]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

from edustudio.utils.common import PathUtil as pathUtil
from edustudio.utils.common import IDUtil as idUtil
import os
from edustudio import __version__

ID = idUtil.get_random_id_bytime() # RUN ID
EDUSTUDIO_VERSION = __version__

WORK_DIR = os.getcwd()

DATA_FOLDER_PATH = f"{WORK_DIR}/data"
TEMP_FOLDER_PATH = f"{WORK_DIR}/temp"
ARCHIVE_FOLDER_PATH = f"{WORK_DIR}/archive"
CFG_FOLDER_PATH = f"{WORK_DIR}/conf"

pathUtil.auto_create_folder_path(
    TEMP_FOLDER_PATH,
    ARCHIVE_FOLDER_PATH,
    DATA_FOLDER_PATH,
    CFG_FOLDER_PATH,
)

DISABLE_TQDM_BAR = False
LOG_WITHOUT_DATE = False
TQDM_NCOLS = 100
DISABLE_LOG_STDOUT = False

curr_file_folder = os.path.dirname(__file__)
DT_INFO_FILE_PATH = os.path.realpath(curr_file_folder + "/assets/datasets.yaml")

DT_INFO_DICT = {} # additional dataset info entrypoint, example: {'FrcSub': {middata_url: https://huggingface.co/datasets/lmcRS/edustudio-datasets/resolve/main/FrcSub/FrcSub-middata.zip} }



r"""
GKT
##########################################

Reference:
    Hiromi Nakagawa et al. "Graph-based Knowledge Tracing: Modeling Student Proficiency Using Graph Neural Network" in WI 2019.

Reference Code:
    https://github.com/jhljx/GKT

"""

from ...gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F
import numpy as np
from .building_blocks import MLP, MLPDecoder, MLPEncoder, EraseAddGate, ScaledDotProductAttention
import scipy.sparse as sp
from .losses import VAELoss, gumbel_softmax


class GKT(GDBaseModel):
    default_cfg = {
        'emb_dim': 32,
        'hidden_dim': 32,
        'edge_type_num': 2,
        'graph_type': 'Dense',
        'dropout': 0.0,
        'MHA_cfg': {
            'atten_dim': 32,
        },
        'VAE_cfg': {
            'encoder_dim': 32,
            'decoder_dim': 32,
            'temp': 0.5,
            'bias': True,
            'factor': False,
            'var': 1.0
        }
    }

    def add_extra_data(self, **kwargs):
        self.graph = kwargs['graph']

    def build_cfg(self):
        self.graph_type = self.modeltpl_cfg['graph_type']
        self.n_stu = self.datatpl_cfg['dt_info']['stu_count']
        self.n_exer = self.datatpl_cfg['dt_info']['exer_count']
        self.edge_type_num = self.modeltpl_cfg['edge_type_num']

        self.hidden_dim = self.modeltpl_cfg['hidden_dim']
        self.emb_dim = self.modeltpl_cfg['emb_dim']

        assert self.graph_type in ['Dense', 'Transition', 'DKT', 'PAM', 'MHA', 'VAE']

    def build_model(self):
        self.exer_emb = nn.Embedding(self.n_exer * 2, self.emb_dim)
        self.cpt_emb = nn.Embedding(self.n_exer, self.emb_dim)

        if self.graph_type in ['Dense', 'Transition', 'DKT']:
            assert  self.edge_type_num == 2
            assert self.graph is not None
            self.graph = nn.Parameter(self.graph)  # [concept_num, concept_num]
            self.graph.requires_grad = False  # fix parameter
        else:  # ['PAM', 'MHA', 'VAE']
            if self.graph_type == 'PAM':
                assert self.graph is None
                self.graph = nn.Parameter(torch.rand(self.n_exer, self.n_exer))
            elif self.graph_type == 'MHA':
                self.graph_model = MultiHeadAttention(
                    2, self.n_exer, self.emb_dim,
                    self.modeltpl_cfg['MHA_cfg']['atten_dim'],
                    dropout=self.modeltpl_cfg['dropout']
                )
            elif self.graph_type == 'VAE':
                vae_cfg = self.modeltpl_cfg['VAE_cfg']
                self.graph_model = VAE(
                    self.emb_dim, vae_cfg['encoder_dim'], 2, vae_cfg['decoder_dim'], vae_cfg['decoder_dim'], 
                    self.n_exer, edge_type_num=2, tau=vae_cfg['temp'], 
                    factor=vae_cfg['factor'], dropout=self.modeltpl_cfg['dropout'], bias=vae_cfg['bias']
                )
                self.vae_loss = VAELoss(self.n_exer, 2, prior=False, var=vae_cfg['var'])
            else:
                raise ValueError(f"unknown graph_type: {self.graph_type}")
        dropout = self.modeltpl_cfg['dropout']
        mlp_input_dim =  self.hidden_dim + self.emb_dim

        self.f_self = MLP(mlp_input_dim, self.hidden_dim, self.hidden_dim, dropout=dropout)
        self.f_neighbor_list = nn.ModuleList()
        if self.graph_type in ['Dense', 'Transition', 'DKT', 'PAM']:
            # f_in and f_out functions
            self.f_neighbor_list.append(MLP(2 * mlp_input_dim, self.hidden_dim, self.hidden_dim, dropout=dropout))
            self.f_neighbor_list.append(MLP(2 * mlp_input_dim, self.hidden_dim, self.hidden_dim, dropout=dropout))
        else:  # ['MHA', 'VAE']
            for _ in range(self.edge_type_num):
                self.f_neighbor_list.append(MLP(2 * mlp_input_dim, self.hidden_dim, self.hidden_dim, dropout=dropout))

        # Erase & Add Gate
        self.erase_add_gate = EraseAddGate(self.hidden_dim, self.n_exer)
        # Gate Recurrent Unit
        self.gru = nn.GRUCell(self.hidden_dim, self.hidden_dim, bias=True)
        # prediction layer
        self.pd_layer = nn.Linear(self.hidden_dim, 1, bias=True)

    # Aggregate step, as shown in Section 3.2.1 of the paper
    def _aggregate(self, xt, qt, ht, qt_mask):
        res_embedding = self.exer_emb(xt)
        concept_embedding = self.cpt_emb.weight.repeat(qt.shape[0], 1).reshape(qt.shape[0], self.n_exer, -1) # [batch_size, concept_num, embedding_dim]
        index_tuple = (torch.arange(qt_mask.sum(), device=self.device), qt[qt_mask].long())
        concept_embedding[qt_mask] = concept_embedding[qt_mask].index_put(index_tuple, res_embedding[qt_mask]) # 只对自身进行替换，替换成另一种embebdding
        tmp_ht = torch.cat((ht, concept_embedding), dim=-1)  # [batch_size, concept_num, hidden_dim + embedding_dim]
        return tmp_ht

    # GNN aggregation step, as shown in 3.3.2 Equation 1 of the paper
    def _agg_neighbors(self, tmp_ht, qt, qt_mask):
        masked_qt = qt[qt_mask]  # [mask_num, ]
        masked_tmp_ht = tmp_ht[qt_mask]  # [mask_num, concept_num, hidden_dim + embedding_dim]
        mask_num = masked_tmp_ht.shape[0]
        self_index_tuple = (torch.arange(mask_num, device=qt.device), masked_qt.long())
        self_ht = masked_tmp_ht[self_index_tuple]  # [mask_num, hidden_dim + embedding_dim]
        self_features = self.f_self(self_ht)  # [mask_num, hidden_dim]
        expanded_self_ht = self_ht.unsqueeze(dim=1).repeat(1, self.n_exer, 1)  #[mask_num, concept_num, hidden_dim + embedding_dim]
        neigh_ht = torch.cat((expanded_self_ht, masked_tmp_ht), dim=-1)  #[mask_num, concept_num, 2 * (hidden_dim + embedding_dim)]
        concept_embedding, rec_embedding, z_prob = None, None, None

        if self.graph_type in ['Dense', 'Transition', 'DKT', 'PAM']:
            adj = self.graph[masked_qt.long(), :].unsqueeze(dim=-1)  # [mask_num, concept_num, 1]
            reverse_adj = self.graph[:, masked_qt.long()].transpose(0, 1).unsqueeze(dim=-1)  # [mask_num, concept_num, 1]
            # self.f_neighbor_list[0](neigh_ht) shape: [mask_num, concept_num, hidden_dim]
            neigh_features = adj * self.f_neighbor_list[0](neigh_ht) + reverse_adj * self.f_neighbor_list[1](neigh_ht)
        else:  # ['MHA', 'VAE']
            concept_embedding = self.cpt_emb.weight  # [concept_num, embedding_dim]
            if self.graph_type == 'MHA':
                query = self.cpt_emb(masked_qt)
                key = concept_embedding
                att_mask = torch.ones(self.edge_type_num, mask_num, self.n_exer, device=qt.device)
                for k in range(self.edge_type_num):
                    index_tuple = (torch.arange(mask_num, device=qt.device), masked_qt.long())
                    att_mask[k] = att_mask[k].index_put(index_tuple, torch.zeros(mask_num, device=qt.device))
                graphs = self.graph_model(masked_qt, query, key, att_mask)
            else:  # self.graph_type == 'VAE'
                sp_send, sp_rec, sp_send_t, sp_rec_t = self._get_edges(masked_qt)
                graphs, rec_embedding, z_prob = self.graph_model(concept_embedding, sp_send, sp_rec, sp_send_t, sp_rec_t)
            neigh_features = 0
            for k in range(self.edge_type_num):
                adj = graphs[k][masked_qt, :].unsqueeze(dim=-1)  # [mask_num, concept_num, 1]
                if k == 0:
                    neigh_features = adj * self.f_neighbor_list[k](neigh_ht)
                else:
                    neigh_features = neigh_features + adj * self.f_neighbor_list[k](neigh_ht)
            if self.graph_type == 'MHA':
                neigh_features = 1. / self.edge_type_num * neigh_features
        # neigh_features: [mask_num, concept_num, hidden_dim]
        m_next = tmp_ht[:, :, :self.hidden_dim]
        m_next[qt_mask] = neigh_features
        m_next[qt_mask] = m_next[qt_mask].index_put(self_index_tuple, self_features)
        return m_next, concept_embedding, rec_embedding, z_prob

    # Update step, as shown in Section 3.3.2 of the paper
    def _update(self, tmp_ht, ht, qt, qt_mask):
        r"""
        Parameters:
            tmp_ht: temporal hidden representations of all concepts after the aggregate step
            ht: hidden representations of all concepts at the current timestamp
            qt: question indices for all students in a batch at the current timestamp
        Shape:
            tmp_ht: [batch_size, concept_num, hidden_dim + embedding_dim]
            ht: [batch_size, concept_num, hidden_dim]
            qt: [batch_size]
            h_next: [batch_size, concept_num, hidden_dim]
        Return:
            h_next: hidden representations of all concepts at the next timestamp
            concept_embedding: input of VAE (optional)
            rec_embedding: reconstructed input of VAE (optional)
            z_prob: probability distribution of latent variable z in VAE (optional)
        """
        mask_num = qt_mask.nonzero().shape[0]
        # GNN Aggregation
        m_next, concept_embedding, rec_embedding, z_prob = self._agg_neighbors(tmp_ht, qt, qt_mask)  # [batch_size, concept_num, hidden_dim]
        # Erase & Add Gate
        m_next[qt_mask] = self.erase_add_gate(m_next[qt_mask])  # [mask_num, concept_num, hidden_dim]
        # GRU
        h_next = m_next
        res = self.gru(m_next[qt_mask].reshape(-1, self.hidden_dim), ht[qt_mask].reshape(-1, self.hidden_dim))  # [mask_num * concept_num, hidden_num]
        index_tuple = (torch.arange(mask_num, device=qt_mask.device), )
        h_next[qt_mask] = h_next[qt_mask].index_put(index_tuple, res.reshape(-1, self.n_exer, self.hidden_dim))
        return h_next, concept_embedding, rec_embedding, z_prob

    # Predict step, as shown in Section 3.3.3 of the paper
    def _predict(self, h_next, qt):
        r"""
        Parameters:
            h_next: hidden representations of all concepts at the next timestamp after the update step
            qt: question indices for all students in a batch at the current timestamp
        Shape:
            h_next: [batch_size, concept_num, hidden_dim]
            qt: [batch_size]
            y: [batch_size, concept_num]
        Return:
            y: predicted correct probability of all concepts at the next timestamp
        """
        qt_mask = torch.ne(qt, -1)  # [batch_size], qt != -1
        y = self.pd_layer(h_next).squeeze(dim=-1)  # [batch_size, concept_num]
        y[qt_mask] = torch.sigmoid(y[qt_mask])  # [batch_size, concept_num]
        return y

    def _get_next_pred(self, yt, q_next):
        r"""
        Parameters:
            yt: predicted correct probability of all concepts at the next timestamp
            q_next: question index matrix at the next timestamp
            batch_size: the size of a student batch
        Shape:
            y: [batch_size, concept_num]
            questions: [batch_size, seq_len]
            pred: [batch_size, ]
        Return:
            pred: predicted correct probability of the question answered at the next timestamp
        """
        return yt.gather(index=q_next.reshape(-1,1), dim=1)

    # Get edges for edge inference in VAE
    def _get_edges(self, masked_qt):
        r"""
        Parameters:
            masked_qt: qt index with -1 padding values removed
        Shape:
            masked_qt: [mask_num, ]
            rel_send: [edge_num, concept_num]
            rel_rec: [edge_num, concept_num]
        Return:
            rel_send: from nodes in edges which send messages to other nodes
            rel_rec:  to nodes in edges which receive messages from other nodes
        """
        mask_num = masked_qt.shape[0]
        row_arr = masked_qt.cpu().numpy().reshape(-1, 1)  # [mask_num, 1]
        row_arr = np.repeat(row_arr, self.n_exer, axis=1)  # [mask_num, concept_num]
        col_arr = np.arange(self.n_exer).reshape(1, -1)  # [1, concept_num]
        col_arr = np.repeat(col_arr, mask_num, axis=0)  # [mask_num, concept_num]
        # add reversed edges
        new_row = np.vstack((row_arr, col_arr))  # [2 * mask_num, concept_num]
        new_col = np.vstack((col_arr, row_arr))  # [2 * mask_num, concept_num]
        row_arr = new_row.flatten()  # [2 * mask_num * concept_num, ]
        col_arr = new_col.flatten()  # [2 * mask_num * concept_num, ]
        data_arr = np.ones(2 * mask_num * self.n_exer)
        init_graph = sp.coo_matrix((data_arr, (row_arr, col_arr)), shape=(self.n_exer, self.n_exer))
        init_graph.setdiag(0)  # remove self-loop edges
        row_arr, col_arr, _ = sp.find(init_graph)
        row_tensor = torch.from_numpy(row_arr).long()
        col_tensor = torch.from_numpy(col_arr).long()
        one_hot_table = torch.eye(self.n_exer, self.n_exer)
        rel_send = F.embedding(row_tensor, one_hot_table)  # [edge_num, concept_num]
        rel_rec = F.embedding(col_tensor, one_hot_table)  # [edge_num, concept_num]
        sp_rec, sp_send = rel_rec.to_sparse(), rel_send.to_sparse()
        sp_rec_t, sp_send_t = rel_rec.T.to_sparse(), rel_send.T.to_sparse()
        sp_send = sp_send.to(device=masked_qt.device)
        sp_rec = sp_rec.to(device=masked_qt.device)
        sp_send_t = sp_send_t.to(device=masked_qt.device)
        sp_rec_t = sp_rec_t.to(device=masked_qt.device)
        return sp_send, sp_rec, sp_send_t, sp_rec_t

    def forward(self, exer_seq, label_seq, mask_seq, **kwargs):
        batch_size, seq_len = exer_seq.shape
        features = exer_seq + label_seq.long() * self.n_exer

        ht = torch.zeros((batch_size, self.n_exer, self.hidden_dim), device=self.device)
        pred_list = []
        ec_list = []  # concept embedding list in VAE
        rec_list = []  # reconstructed embedding list in VAE
        z_prob_list = []  # probability distribution of latent variable z in VAE
        for i in range(seq_len):
            xt = features[:, i]  # [batch_size]
            qt = exer_seq[:, i]  # [batch_size]
            qt_mask = torch.eq(mask_seq[:, i], 1)  # [batch_size], next_qt != -1
            tmp_ht = self._aggregate(xt, qt, ht, qt_mask)  # [batch_size, concept_num, hidden_dim + embedding_dim]
            h_next, concept_embedding, rec_embedding, z_prob = self._update(tmp_ht, ht, qt, qt_mask)  # [batch_size, concept_num, hidden_dim]
            ht[qt_mask] = h_next[qt_mask]  # update new ht
            yt = self._predict(h_next, qt)  # [batch_size, concept_num]
            if i < seq_len - 1:
                pred = self._get_next_pred(yt, exer_seq[:, i + 1], )
                pred_list.append(pred)
            ec_list.append(concept_embedding)
            rec_list.append(rec_embedding)
            z_prob_list.append(z_prob)
        pred_res = torch.stack(pred_list, dim=1)  # [batch_size, seq_len - 1]
        return pred_res, ec_list, rec_list, z_prob_list
    
    def get_main_loss(self, **kwargs):
        y_pd, ec_list, rec_list, z_prob_list = self(**kwargs)
        y_pd = y_pd.squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )

        if self.graph_type != 'VAE':
            return {
                'loss_main': loss,
            }
        else:
            return {
                'loss_main': loss,
                'loss_vae': self.vae_loss(ec_list, rec_list, z_prob_list)
            }


    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)
    
    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)[0].squeeze(dim=-1)
        # y_pd = y_pd.gather(
        #     index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=2
        # ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }
    

class MultiHeadAttention(nn.Module):
    """
    Multi-Head Attention module
    NOTE: Stole and modify from https://github.com/jadore801120/attention-is-all-you-need-pytorch/blob/master/transformer/SubLayers.py
    """

    def __init__(self, n_head, concept_num, input_dim, d_k, dropout=0.):
        super(MultiHeadAttention, self).__init__()
        self.n_head = n_head
        self.n_exer = concept_num
        self.d_k = d_k
        self.w_qs = nn.Linear(input_dim, n_head * d_k, bias=False)
        self.w_ks = nn.Linear(input_dim, n_head * d_k, bias=False)
        self.attention = ScaledDotProductAttention(temperature=d_k ** 0.5, attn_dropout=dropout)
        # inferred latent graph, used for saving and visualization
        self.graphs = nn.Parameter(torch.zeros(n_head, concept_num, concept_num))
        self.graphs.requires_grad = False

    def _get_graph(self, attn_score, qt):
        r"""
        Parameters:
            attn_score: attention score of all queries
            qt: masked question index
        Shape:
            attn_score: [n_head, mask_num, concept_num]
            qt: [mask_num]
        Return:
            graphs: n_head types of inferred graphs
        """
        graphs = torch.zeros(self.n_head, self.n_exer, self.n_exer, device=qt.device)
        for k in range(self.n_head):
            index_tuple = (qt.long(), )
            graphs[k] = graphs[k].index_put(index_tuple, attn_score[k])  # used for calculation
            #############################
            # here, we need to detach edges when storing it into self.graphs in case memory leak!
            self.graphs.data[k] = self.graphs.data[k].index_put(index_tuple, attn_score[k].detach())  # used for saving and visualization
            #############################
        return graphs

    def forward(self, qt, query, key, mask=None):
        r"""
        Parameters:
            qt: masked question index
            query: answered concept embedding for a student batch
            key: concept embedding matrix
            mask: mask matrix
        Shape:
            qt: [mask_num]
            query: [mask_num, embedding_dim]
            key: [concept_num, embedding_dim]
        Return:
            graphs: n_head types of inferred graphs
        """
        d_k, n_head = self.d_k, self.n_head
        len_q, len_k = query.size(0), key.size(0)

        # Pass through the pre-attention projection: lq x (n_head *dk)
        # Separate different heads: lq x n_head x dk
        q = self.w_qs(query).view(len_q, n_head, d_k)
        k = self.w_ks(key).view(len_k, n_head, d_k)

        # Transpose for attention dot product: n_head x lq x dk
        q, k = q.transpose(0, 1), k.transpose(0, 1)
        attn_score = self.attention(q, k, mask=mask)  # [n_head, mask_num, concept_num]
        graphs = self._get_graph(attn_score, qt)
        return graphs


class VAE(nn.Module):

    def __init__(self, input_dim, hidden_dim, output_dim, msg_hidden_dim, msg_output_dim, concept_num, edge_type_num=2,
                 tau=0.1, factor=True, dropout=0., bias=True):
        super(VAE, self).__init__()
        self.edge_type_num = edge_type_num
        self.n_exer = concept_num
        self.tau = tau
        self.encoder = MLPEncoder(input_dim, hidden_dim, output_dim, factor=factor, dropout=dropout, bias=bias)
        self.decoder = MLPDecoder(input_dim, msg_hidden_dim, msg_output_dim, hidden_dim, edge_type_num, dropout=dropout, bias=bias)
        # inferred latent graph, used for saving and visualization
        self.graphs = nn.Parameter(torch.zeros(edge_type_num, concept_num, concept_num))
        self.graphs.requires_grad = False

    def _get_graph(self, edges, sp_rec, sp_send):
        r"""
        Parameters:
            edges: sampled latent graph edge weights from the probability distribution of the latent variable z
            sp_rec: one-hot encoded receive-node index(sparse tensor)
            sp_send: one-hot encoded send-node index(sparse tensor)
        Shape:
            edges: [edge_num, edge_type_num]
            sp_rec: [edge_num, concept_num]
            sp_send: [edge_num, concept_num]
        Return:
            graphs: latent graph list modeled by z which has different edge types
        """
        x_index = sp_send._indices()[1].long()  # send node index: [edge_num, ]
        y_index = sp_rec._indices()[1].long()   # receive node index [edge_num, ]
        graphs = torch.zeros(self.edge_type_num, self.n_exer, self.n_exer, device=edges.device)
        for k in range(self.edge_type_num):
            index_tuple = (x_index, y_index)
            graphs[k] = graphs[k].index_put(index_tuple, edges[:, k])  # used for calculation
            #############################
            # here, we need to detach edges when storing it into self.graphs in case memory leak!
            self.graphs.data[k] = self.graphs.data[k].index_put(index_tuple, edges[:, k].detach())  # used for saving and visualization
            #############################
        return graphs

    def forward(self, data, sp_send, sp_rec, sp_send_t, sp_rec_t):
        r"""
        Parameters:
            data: input concept embedding matrix
            sp_send: one-hot encoded send-node index(sparse tensor)
            sp_rec: one-hot encoded receive-node index(sparse tensor)
            sp_send_t: one-hot encoded send-node index(sparse tensor, transpose)
            sp_rec_t: one-hot encoded receive-node index(sparse tensor, transpose)
        Shape:
            data: [concept_num, embedding_dim]
            sp_send: [edge_num, concept_num]
            sp_rec: [edge_num, concept_num]
            sp_send_t: [concept_num, edge_num]
            sp_rec_t: [concept_num, edge_num]
        Return:
            graphs: latent graph list modeled by z which has different edge types
            output: the reconstructed data
            prob: q(z|x) distribution
        """
        logits = self.encoder(data, sp_send, sp_rec, sp_send_t, sp_rec_t)  # [edge_num, output_dim(edge_type_num)]
        edges = gumbel_softmax(logits, tau=self.tau, dim=-1)  # [edge_num, edge_type_num]
        prob = F.softmax(logits, dim=-1)
        output = self.decoder(data, edges, sp_send, sp_rec, sp_send_t, sp_rec_t)  # [concept_num, embedding_dim]
        graphs = self._get_graph(edges, sp_send, sp_rec)
        return graphs, output, prob



import pandas as pd
import time
from .raw2mid import BaseRaw2Mid

r"""
R2M_Bridge2Algebra_0607
########################
"""


class R2M_Bridge2Algebra_0607(BaseRaw2Mid):
    """R2M_Bridge2Algebra_0607 is a class used to handle the Bridge2Algebra 2006-2007 dataset."""

    def process(self):
        super().process()
        pd.set_option("mode.chained_assignment", None)  # ignore warning
        # 读取数据集，并展示相关数据
        raw_data_train = pd.read_table(f"{self.rawpath}/bridge_to_algebra_2006_2007_train.txt", encoding='utf-8',
                                       low_memory=False)
        raw_data_test = pd.read_table(f"{self.rawpath}/bridge_to_algebra_2006_2007_master.txt", encoding='utf-8',
                                      low_memory=False)
        raw_data = pd.concat([raw_data_train, raw_data_test], axis=0)
        raw_data.head()

        # 预处理信息
        # 1.去除空值
        data = raw_data.copy().dropna(subset=['Step Start Time'])
        # 2. 时间格式转换为时间戳, 单位是ms
        data['Step Start Time'] = data['Step Start Time'].apply(
            lambda x: int(time.mktime(time.strptime(str(x)[:-2], "%Y-%m-%d %H:%M:%S"))) * 1000)

        # 以'Anon Student Id', 'Problem Name'为分组依据，将学生处理每个习题的步骤合并到一起
        # 同时处理一些数据，包括assignment_name，cost_time，start_timestamp，correct
        def concat_func(x):
            return pd.Series({
                'assignment_name': x['Problem Hierarchy'].unique(),
                'cost_time': x['Step Duration (sec)'].sum(),
                'steps': ','.join(x['Step Name']),
                'start_timestamp': x['Step Start Time'].min(),
                'correct': x['Corrects'].min(),  # 我们认为只要学生步骤中有一步没有作对，那么整个题就不对
                'KC(SubSkills)': '--'.join(str(id) for id in x['KC(SubSkills)'])  # 遇到none怎么办？
            })

        raw_data = data.groupby(['Anon Student Id', 'Problem Name'], as_index=False).apply(concat_func)

        # 定义一个name_to_id的函数，返回值为两个dict,一个是name2id，一个是id2name
        # 主要用于将唯一标识的字符串生成自增的id
        def name_to_id(l: list()):
            if l is None:
                return None, None
            name2id = dict(zip(l, range(len(l))))
            id2name = dict(zip(range(len(l)), l))
            return name2id, id2name

        # 获取交互信息
        # 若用户正确次数超过一次，则认为用户已经正确做答
        def correct(corrects):
            if corrects >= 1:
                return 1
            return 0

        inter = pd.DataFrame.copy(raw_data)
        # 接下来就添加label、order、cost_time信息
        inter['label'] = inter['correct'].apply(lambda x: correct(x))
        # 根据时间戳确定学生做题的序列
        inter = inter.sort_values(by='start_timestamp', ascending=True)
        inter['order'] = range(len(inter))
        # cost_time 转化为以ms为单位
        inter['cost_time'] = inter['cost_time'].apply(lambda x: int(x * 1000))
        # 先重命名，方便后续操作
        df_inter = inter.rename(columns={'Anon Student Id': 'stu_name', 'Problem Name': 'exer_name'})

        # 处理name to id的映射，及最终df_inter的标准格式
        # 处理id映射，包括将字符类型的stu_name,exer_name映射到数字类型的stu_id和exer_id
        unique_stu = list(df_inter['stu_name'].copy().unique())
        unique_exer = list(df_inter['exer_name'].copy().unique())
        # 设置stu2id,和id2stu的dict及exer2id和id2exer的dict
        stu2id, id2stu = name_to_id(unique_stu)
        exer2id, id2exer = name_to_id(unique_exer)
        df_inter['stu_id'] = df_inter['stu_name'].apply(lambda x: stu2id[x])
        df_inter['exer_id'] = df_inter['exer_name'].apply(lambda x: exer2id[x])
        # 最终处理，重排列序
        df_inter = df_inter.reindex(
            columns=['stu_id', 'exer_id', 'label', 'start_timestamp', 'cost_time', 'order']).rename(
            columns={'stu_id': 'stu_id:token', 'exer_id': 'exer_id:token', 'label': 'label:float',
                     'start_timestamp': 'start_timestamp:float', 'cost_time': 'cost_time:float',
                     'order': 'order_id:float'})

        # 处理用户信息
        df_stu = df_inter['stu_id:token'].copy().unique()
        # df_stu['classId'] = None
        # df_stu['gender'] = None

        # 处理习题信息
        # kcs是一个set集合，包括整个训练集的所有cpt信息;ks为每个习题所关联的cpt_seq信息
        # cpt2id 是 cpt_name 到 cpt_id 的字典;id2cpt 是 cpt_id 到 cpt_name 的字典
        kcs = set()
        cpt2id = dict()
        id2cpt = dict()

        # 对KC(Default)进行预处理，将其转化cpt_seq_name序列集合
        def preprocess_cpt_seq(s):
            cpt_seq_name = set()
            if type(s) == str:
                kc = s.split("--")
                for i in kc:
                    if i[0] == '[':
                        i = [s.strip("- {}]") for s in i[12:].split(";")]
                        cpt_seq_name.update(i)
                    else:
                        cpt_seq_name.add(i)
            if 'nan' in cpt_seq_name:  # 去除空值nan
                cpt_seq_name.remove('nan')
            # cpt_seq_name.pop('nan')
            kcs.update(cpt_seq_name)
            if len(cpt_seq_name) == 0:
                cpt_seq_name = None
            return cpt_seq_name

        # 将cpt_seq_name的字符串集合转化为cpt_seq的数字列表
        def process_cpt_seq(cpt_seq_name: set()):
            if cpt_seq_name is None:
                return None
            cpt_seq = [cpt2id[cpt_name] for cpt_name in cpt_seq_name]
            cpt_seq.sort()
            return cpt_seq

        # 对exercise的处理,主要将KC(Default)处理成cpt_seq_name序列
        exer = pd.DataFrame.copy(raw_data).drop_duplicates(subset=['Problem Name'])
        exer['exer_id'] = exer['Problem Name'].apply(lambda x: exer2id[x])
        # 将字符串形式的assignment转化为数字形式的assignment_id
        unique_assignment = list(exer['assignment_name'].apply(lambda x: x[0]).unique())
        assignment2id, id2assignment = name_to_id(unique_assignment)
        exer['assignment_id'] = exer['assignment_name'].apply(lambda x: assignment2id[x[0]])
        # 将每个习题的KC(Default)->cpt_seq_name->cpt_seq，
        exer['cpt_seq_name'] = exer.apply(lambda x: preprocess_cpt_seq(x['KC(SubSkills)']), axis=1)
        cpt2id, id2cpt = name_to_id(kcs)
        exer['cpt_seq'] = exer['cpt_seq_name'].apply(lambda x: process_cpt_seq(x))
        df_exer = exer.reindex(columns=['exer_id', 'assignment_id', 'cpt_seq']).rename(
            columns={'exer_id': 'exer_id:token', 'assignment_id': 'assignment_id:token_seq',
                     'cpt_seq': 'cpt_seq:token_seq'})

        # 此处将数据保存到`cfg.MIDDATA_PATH`中
        df_inter.to_csv(f"{self.midpath}/{self.dt}.inter.csv", index=False, encoding='utf-8')
        # df_stu.to_csv(f"{self.midpath}/{self.dt}.stu.csv", index=False, encoding='utf-8')
        df_exer.to_csv(f"{self.midpath}/{self.dt}.exer.csv", index=False, encoding='utf-8')
        pd.set_option("mode.chained_assignment", "warn")  # ignore warning
        return



from ..common.base_mid2cache import BaseMid2Cache
import pandas as pd
import numpy as np
from edustudio.datatpl.utils import SpliterUtil, PadSeqUtil
from itertools import chain


class M2C_BuildSeqInterFeats(BaseMid2Cache):
    default_cfg = {
        'window_size': 100,
        "extra_inter_feats": []
    }
    
    def __init__(self, m2c_cfg, n_folds, is_dataset_divided) -> None:
        super().__init__(m2c_cfg)
        self.n_folds = n_folds
        self.is_dataset_divided = is_dataset_divided

    @classmethod
    def from_cfg(cls, cfg):
        m2c_cfg = cfg.datatpl_cfg.get(cls.__name__)
        n_folds = cfg.datatpl_cfg.n_folds
        is_dataset_divided = cfg.datatpl_cfg.is_dataset_divided
        return cls(m2c_cfg, n_folds, is_dataset_divided)
    
    def process(self, **kwargs):
        df = kwargs['df']
        df_train, df_valid, df_test = kwargs['df_train'], kwargs['df_valid'], kwargs['df_test']
        df = self.sort_records(df)
        df_train = self.sort_records(df_train)
        df_valid = self.sort_records(df_valid)
        df_test = self.sort_records(df_test)

        if not self.is_dataset_divided:
            assert df_train is None and df_valid is None and df_test is None
            self.window_size = self.m2c_cfg['window_size']
            if self.m2c_cfg['window_size'] <= 0 or self.m2c_cfg['window_size'] is None:
                self.window_size = df[['stu_id:token', 'exer_id:token']].groupby('stu_id:token').agg('count')['exer_id:token'].max()
            self.logger.info(f"actual window size: {self.window_size}")
            kwargs['df_seq'] = self.construct_df2dict(df)

        else: # dataset is divided
            assert df_train is not None and df_test is not None
            if self.m2c_cfg['window_size'] <= 0 or self.m2c_cfg['window_size'] is None:
                self.window_size = np.max([
                    df_train[['stu_id:token', 'exer_id:token']].groupby('stu_id:token').agg('count')['exer_id:token'].max(),
                    df_valid[['stu_id:token', 'exer_id:token']].groupby('stu_id:token').agg('count')['exer_id:token'].max() if df_valid is not None else 0,
                    df_test[['stu_id:token', 'exer_id:token']].groupby('stu_id:token').agg('count')['exer_id:token'].max()
                ])
            else:
                self.window_size = self.m2c_cfg['window_size']
            self.logger.info(f"actual window size: {self.window_size}")

            train_dict = self.construct_df2dict(df_train)
            valid_dict = self.construct_df2dict(df_valid)
            test_dict = self.construct_df2dict(df_test)
            kwargs['df_train_seq'] = train_dict
            kwargs['df_valid_seq'] = valid_dict
            kwargs['df_test_seq'] = test_dict
        return kwargs

    @staticmethod
    def sort_records(df, col=None):
        if df is not None:
            if col is None:
                if 'order_id:float' in df.columns:
                    col = 'order_id:float'
                elif 'order_id:token' in df.columns:
                    col = 'order_id:token'
                else:
                    raise ValueError("please specificy the sort column")

            return df.sort_values(by=col, ascending=True).reset_index(drop=True)
    
    def construct_df2dict(self, df: pd.DataFrame):
        if df is None: return None

        tmp_df = df[['stu_id:token','exer_id:token','label:float'] + self.m2c_cfg['extra_inter_feats']].groupby('stu_id:token').agg(lambda x: list(x)).reset_index()

        exer_seq, idx, mask_seq = PadSeqUtil.pad_sequence(
            tmp_df['exer_id:token'].to_list(), return_idx=True, return_mask=True, 
            maxlen=self.window_size
        )
        label_seq, _, _ = PadSeqUtil.pad_sequence(
            tmp_df['label:float'].to_list(), dtype=np.float32,
            maxlen=self.window_size
        )
        stu_id = tmp_df['stu_id:token'].to_numpy()[idx]

        ret_dict = {
            'stu_id:token': stu_id,
            'exer_seq:token_seq': exer_seq,
            'label_seq:float_seq': label_seq,
            'mask_seq:token_seq': mask_seq
        }
        for extra_feat in self.m2c_cfg['extra_inter_feats']:
            name, type_ = extra_feat.split(":")
            if type_ == 'token':
                seq, _, _ = PadSeqUtil.pad_sequence(
                    tmp_df[extra_feat].to_list(), dtype=np.int64,
                    maxlen=self.window_size
                )
                ret_dict[f"{name}_seq:token_seq"] = seq
            elif type_ == 'float':
                seq, _, _ = PadSeqUtil.pad_sequence(
                    tmp_df[extra_feat].to_list(), dtype=np.float32,
                    maxlen=self.window_size
                )
                ret_dict[f"{name}_seq:float_seq"] = seq
            else:
                raise NotImplementedError

        return ret_dict



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias


from .commonUtil import PathUtil, IDUtil, tensor2npy, tensor2cpu, IOUtil, set_same_seeds, get_gpu_usage, DecoratorTimer
from .configUtil import UnifyConfig, NumpyEncoder
from .loggerUtil import Logger



from .base_mid2cache import BaseMid2Cache
import numpy as np


class M2C_Label2Int(BaseMid2Cache):
    def process(self, **kwargs):
        self.op_on_df('df', kwargs)
        self.op_on_df('df_train', kwargs)
        self.op_on_df('df_valid', kwargs)
        self.op_on_df('df_test', kwargs)
        return kwargs

    @staticmethod
    def op_on_df(column, kwargs):
        if column in kwargs and kwargs[column] is not None:
            kwargs[column]['label:float'] =  (kwargs[column]['label:float'] >= 0.5).astype(np.float32)



from ..common import EduDataTPL


class LPKTDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_ReMapId', 'M2C_BuildSeqInterFeats', 'M2C_RandomDataSplit4KT', 'M2C_LPKT_OP', "M2C_GenQMat"],
        'M2C_BuildSeqInterFeats': {
            "extra_inter_feats": ['start_timestamp:float', 'answer_time:float']
        }
    }

    def set_info_for_fold(self, fold_id):
        dt_info = self.datatpl_cfg['dt_info']
        dt_info['answer_time_count'] = dt_info['answer_time_count_list'][fold_id]
        dt_info['interval_time_count'] = dt_info['interval_time_count_list'][fold_id]



from edustudio.traintpl import GeneralTrainTPL
from edustudio.utils.common import tensor2npy
import torch
from tqdm import tqdm
from edustudio.utils.callback import ModelCheckPoint, EarlyStopping, History, BaseLogger, Callback, CallbackList
from edustudio.utils.common import set_same_seeds
import shutil
import numpy as np

class DCDTrainTPL(GeneralTrainTPL):
    default_cfg = {}

    def one_fold_start(self, fold_id):
        self.logger.info(f"====== [FOLD ID]: {fold_id} ======")
        # callbacks
        num_stop_rounds = self.traintpl_cfg['num_stop_rounds']
        es_metrics = self.traintpl_cfg["early_stop_metrics"]
        modelCheckPoint = ModelCheckPoint(
            es_metrics, save_folder_path=f"{self.frame_cfg.temp_folder_path}/pths/{fold_id}/"
        )
        es_cb = EarlyStopping(es_metrics, num_stop_rounds=num_stop_rounds, start_round=1)
        history_cb = History(folder_path=f"{self.frame_cfg.temp_folder_path}/history/{fold_id}", plot_curve=False)
        callbacks = [
            modelCheckPoint, es_cb, history_cb,
            BaseLogger(self.logger, group_by_contains=['loss', 'user_', 'item_'])
        ]
        self.callback_list = CallbackList(callbacks=callbacks, model=self.model, logger=self.logger)
        # evaltpls
        for evaltpl in self.evaltpls: 
            evaltpl.set_callback_list(self.callback_list)
            evaltpl.set_dataloaders(train_loader=self.train_loader, 
                                    valid_loader=self.valid_loader, 
                                    test_loader=self.test_loader
                                    )
        # train
        # set_same_seeds(self.traintpl_cfg['seed'])
        if type(self.datatpl.filling_Q_mat) is np.ndarray:
            self.model.Q_mat = torch.from_numpy(self.datatpl.filling_Q_mat).to(self.traintpl_cfg.device)
        else:
            self.model.Q_mat = self.datatpl.filling_Q_mat.to(self.traintpl_cfg.device)
        self.model.callback_list = self.callback_list
        if self.valid_loader is not None:
            self.fit(train_loader=self.train_loader, valid_loader=self.valid_loader)
        else:
            self.fit(train_loader=self.train_loader, valid_loader=self.test_loader)
        
        metric_name = self.traintpl_cfg['best_epoch_metric']
        metric = [m for m in modelCheckPoint.metric_list if m.name == metric_name][0]
        if self.valid_loader is not None:
            # load best params
            fpth =  f"{self.frame_cfg.temp_folder_path}/pths/{fold_id}/best-epoch-{metric.best_epoch:03d}-for-{metric.name}.pth"
            self.model.load_state_dict(torch.load(fpth))

            metrics = self.inference(self.test_loader)
            for name in metrics: self.logger.info(f"{name}: {metrics[name]}")
            History.dump_json(metrics, f"{self.frame_cfg.temp_folder_path}/{fold_id}/result.json")
        else:
            # metrics = history_cb.log_as_time[metric.best_epoch]
            # load best params
            fpth =  f"{self.frame_cfg.temp_folder_path}/pths/{fold_id}/best-epoch-{metric.best_epoch:03d}-for-{metric.name}.pth"
            self.model.load_state_dict(torch.load(fpth))

            metrics = self.inference(self.test_loader)
            for name in metrics: self.logger.info(f"{name}: {metrics[name]}")
            History.dump_json(metrics, f"{self.frame_cfg.temp_folder_path}/{fold_id}/result.json")

        if self.traintpl_cfg['unsave_best_epoch_pth']: shutil.rmtree(f"{self.frame_cfg.temp_folder_path}/pths/")
        return metrics
    
    @torch.no_grad()
    def evaluate(self, loader):
        self.model.eval()
        pd_list = list(range(len(loader)))
        gt_list = list(range(len(loader)))
        for idx, batch_dict in enumerate(tqdm(loader, ncols=self.frame_cfg['TQDM_NCOLS'], desc="[PREDICT]")):
            batch_dict = self.batch_dict2device(batch_dict)
            eval_dict = self.model.predict(**batch_dict)
            pd_list[idx] = eval_dict['y_pd']
            gt_list[idx] = eval_dict['y_gt'] if 'y_gt' in eval_dict else batch_dict['label']
        y_pd = torch.hstack(pd_list)
        y_gt = torch.hstack(gt_list)

        eval_data_dict = {
            'y_pd': y_pd,
            'y_gt': y_gt,
        }
        if hasattr(self.model, 'get_stu_status'):
            stu_stats_list = []
            idx = torch.arange(0, self.datatpl_cfg['dt_info']['stu_count']).to(self.traintpl_cfg['device'])
            for i in range(0,self.datatpl_cfg['dt_info']['stu_count'], self.traintpl_cfg['eval_batch_size']):
                batch_stu_id = idx[i:i+self.traintpl_cfg['eval_batch_size']]
                batch = self.model.get_stu_status(batch_stu_id)
                stu_stats_list.append(batch)
            stu_stats = torch.vstack(stu_stats_list)
            eval_data_dict.update({
                'stu_stats': tensor2npy(stu_stats),
            })
        if hasattr(self.model, 'get_exer_emb'):
            eval_data_dict.update({
                'exer_emb': self.model.get_exer_emb(),
            })

        if hasattr(self.datatpl, 'Q_mat'):
            eval_data_dict.update({
                'Q_mat': tensor2npy(self.datatpl.Q_mat)
            })
        eval_result = {}
        for evaltpl in self.evaltpls: eval_result.update(
                evaltpl.eval(ignore_metrics=self.traintpl_cfg['ignore_metrics_in_train'], **eval_data_dict)
            )
        return eval_result


    @torch.no_grad()
    def inference(self, loader):
        self.model.eval()
        pd_list = list(range(len(loader)))
        gt_list = list(range(len(loader)))
        for idx, batch_dict in enumerate(tqdm(loader, ncols=self.frame_cfg['TQDM_NCOLS'], desc="[PREDICT]")):
            batch_dict = self.batch_dict2device(batch_dict)
            eval_dict = self.model.predict(**batch_dict)
            pd_list[idx] = eval_dict['y_pd']
            gt_list[idx] = eval_dict['y_gt'] if 'y_gt' in eval_dict else batch_dict['label']
        y_pd = torch.hstack(pd_list)
        y_gt = torch.hstack(gt_list)

        eval_data_dict = {
            'y_pd': y_pd,
            'y_gt': y_gt,
        }
        if hasattr(self.model, 'get_stu_status'):
            stu_stats_list = []
            idx = torch.arange(0, self.datatpl_cfg['dt_info']['stu_count']).to(self.traintpl_cfg['device'])
            for i in range(0,self.datatpl_cfg['dt_info']['stu_count'], self.traintpl_cfg['eval_batch_size']):
                batch_stu_id = idx[i:i+self.traintpl_cfg['eval_batch_size']]
                batch = self.model.get_stu_status(batch_stu_id)
                stu_stats_list.append(batch)
            stu_stats = torch.vstack(stu_stats_list)
            eval_data_dict.update({
                'stu_stats': tensor2npy(stu_stats),
            })

        if hasattr(self.model, 'get_exer_emb'):
            eval_data_dict.update({
                'exer_emb': self.model.get_exer_emb(),
            })

        if hasattr(self.datatpl, 'Q_mat'):
            eval_data_dict.update({
                'Q_mat': tensor2npy(self.datatpl.Q_mat)
            })
        eval_result = {}
        for evaltpl in self.evaltpls: eval_result.update(evaltpl.eval(**eval_data_dict))
        return eval_result



from ..common.base_mid2cache import BaseMid2Cache
import numpy as np
from edustudio.datatpl.utils import PadSeqUtil


class M2C_GenKCSeq(BaseMid2Cache):
    """Generate Knowledge Component Sequence
    """
    default_cfg = {
        'cpt_seq_window_size': -1,
    }

    def process(self, **kwargs):
        df_exer = kwargs['df_exer']
        tmp_df_Q = df_exer.set_index('exer_id:token')
        exer_count = kwargs['dt_info']['exer_count']
        
        cpt_seq_unpadding = [
            (tmp_df_Q.loc[exer_id].tolist()[0] if exer_id in tmp_df_Q.index else []) for exer_id in range(exer_count)
        ]
        cpt_seq_padding, _, cpt_seq_mask = PadSeqUtil.pad_sequence(
            cpt_seq_unpadding, maxlen=self.m2c_cfg['cpt_seq_window_size'], return_mask=True
        )

        kwargs['cpt_seq_padding'] = cpt_seq_padding
        kwargs['cpt_seq_mask'] = cpt_seq_mask
        return kwargs



r"""
SAINT
##########################################

Reference:
     Youngduck Choi et al. "Towards an Appropriate Query, Key, and Value Computation for Knowledge Tracing" in L-AT-S 2020.

Reference Code:
    https://github.com/pykt-team/pykt-toolkit/blob/main/pykt/models/saint.py

"""
import copy

import math
import pandas as pd
from torch.nn import Dropout

from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F
import numpy as np
from torch.nn.init import xavier_uniform_, constant_

class SAINT(GDBaseModel):
    r"""
    SAINT

    default_cfg:
       'emb_size': 256          # dimension of embedding
        'num_attn_heads': 8     # number of parallel attention heads
        'dropout_rate': 0.2     # dropout rate
        'n_blocks':4          # number of Encoder_blocks
    """
    default_cfg = {
        'emb_size': 256,
        'num_attn_heads': 8,
        'dropout_rate': 0.2,
        'n_blocks':4,
    }


    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.num_q = self.datatpl_cfg['dt_info']['exer_count']
        self.num_c = self.datatpl_cfg['dt_info']['cpt_count']
        self.window_size = self.datatpl_cfg['dt_info']['real_window_size']
        self.dropout_r = self.modeltpl_cfg['dropout_rate']
        self.num_attn_heads = self.modeltpl_cfg['num_attn_heads']
        self.emb_size = self.modeltpl_cfg['emb_size']
        self.n_blocks = self.modeltpl_cfg['n_blocks']
        
    def build_model(self):
        self.embd_pos = nn.Embedding(self.window_size, embedding_dim=self.emb_size)
        self.encoder = get_clones(Encoder_block(self.emb_size, self.num_attn_heads, self.num_q, self.num_c, self.window_size, self.dropout_r, self.device),
                                      self.n_blocks)

        self.decoder = get_clones(Decoder_block(self.emb_size, 2, self.num_attn_heads, self.window_size, self.dropout_r, self.device), self.n_blocks)

        self.dropout = Dropout(self.dropout_r)
        self.out = nn.Linear(in_features=self.emb_size, out_features=1)


    def forward(self, exer_seq, label_seq, cpt_unfold_seq, **kwargs):
        label_seq = label_seq.type(torch.int64)
        if self.num_q > 0:
            in_pos = torch.arange(exer_seq.shape[1]).unsqueeze(0).to(self.device)
        else:
            in_pos = torch.arange(cpt_unfold_seq.shape[1]).unsqueeze(0).to(self.device)
        in_pos = self.embd_pos(in_pos)
        first_block = True
        for i in range(self.n_blocks):
            if i >= 1:
                first_block = False
            exer_seq = self.encoder[i](exer_seq, cpt_unfold_seq, in_pos, first_block=first_block)
            cpt_unfold_seq = exer_seq
        first_block = True
        for i in range(self.n_blocks):
            if i >= 1:
                first_block = False
            label_seq = self.decoder[i](label_seq, in_pos, en_out=exer_seq, first_block=first_block)
        ## Output layer
        res = self.out(self.dropout(label_seq))
        res = torch.sigmoid(res).squeeze(-1)
        return res

    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)


def get_clones(module, N):
    """ Cloning nn modules
    """
    return nn.ModuleList([copy.deepcopy(module) for i in range(N)])

class transformer_FFN(nn.Module):
    def __init__(self, emb_size, dropout) -> None:
        super().__init__()
        self.emb_size = emb_size
        self.dropout = dropout
        self.FFN = nn.Sequential(
                nn.Linear(self.emb_size, self.emb_size),
                nn.ReLU(),
                nn.Dropout(self.dropout),
                nn.Linear(self.emb_size, self.emb_size),
                # Dropout(self.dropout),
            )
    def forward(self, in_fea):
        return self.FFN(in_fea)

def ut_mask(device, seq_len):
    """ Upper Triangular Mask
    """
    return torch.triu(torch.ones(seq_len,seq_len),diagonal=1).to(dtype=torch.bool).to(device)
class Encoder_block(nn.Module):
    """
    M = SkipConct(Multihead(LayerNorm(Qin;Kin;Vin)))
    O = SkipConct(FFN(LayerNorm(M)))
    """

    def __init__(self, dim_model, heads_en, total_ex, total_cat, seq_len, dropout,
                  device, emb_path="", pretrain_dim=768):
        super().__init__()
        self.device = device
        self.seq_len = seq_len
        self.emb_path = emb_path
        self.total_cat = total_cat
        self.total_ex = total_ex
        if total_ex > 0:
            if emb_path == "":
                self.embd_ex = nn.Embedding(total_ex,
                                            embedding_dim=dim_model)  # embedings  q,k,v = E = exercise ID embedding, category embedding, and positionembedding.
            else:
                embs = pd.read_pickle(emb_path)
                self.exercise_embed = nn.Embedding.from_pretrained(embs)
                self.linear = nn.Linear(pretrain_dim, dim_model)
        if total_cat > 0:
            self.emb_cat = nn.Embedding(total_cat, embedding_dim=dim_model)
        # self.embd_pos   = nn.Embedding(seq_len, embedding_dim = dim_model)                  #positional embedding

        self.multi_en = nn.MultiheadAttention(embed_dim=dim_model, num_heads=heads_en, dropout=dropout)
        self.layer_norm1 = nn.LayerNorm(dim_model)
        self.dropout1 = Dropout(dropout)

        self.ffn_en = transformer_FFN(dim_model, dropout)
        self.layer_norm2 = nn.LayerNorm(dim_model)
        self.dropout2 = Dropout(dropout)

    def forward(self, in_ex, in_cat, in_pos, first_block=True):

        ##
        if first_block:
            embs = []
            if self.total_ex > 0:
                if self.emb_path == "":
                    in_ex = self.embd_ex(in_ex)
                else:
                    in_ex = self.linear(self.exercise_embed(in_ex))
                embs.append(in_ex)
            if self.total_cat > 0:
                in_cat = self.emb_cat(in_cat)
                embs.append(in_cat)
            out = embs[0]
            for i in range(1, len(embs)):
                out += embs[i]
            out = out + in_pos
            # in_pos = self.embd_pos(in_pos)
        else:
            out = in_ex

        # in_pos = get_pos(self.seq_len)
        # in_pos = self.embd_pos(in_pos)

        out = out.permute(1, 0, 2)  # (n,b,d)  # print('pre multi', out.shape)

        # norm -> attn -> drop -> skip corresponging to transformers' norm_first
        # Multihead attention
        n, _, _ = out.shape
        out = self.layer_norm1(out)  # Layer norm
        skip_out = out
        out, attn_wt = self.multi_en(out, out, out,
                                     attn_mask=ut_mask(self.device, seq_len=n))  # attention mask upper triangular
        out = self.dropout1(out)
        out = out + skip_out  # skip connection

        # feed forward
        out = out.permute(1, 0, 2)  # (b,n,d)
        out = self.layer_norm2(out)  # Layer norm
        skip_out = out
        out = self.ffn_en(out)
        out = self.dropout2(out)
        out = out + skip_out  # skip connection

        return out


class Decoder_block(nn.Module):
    """
    M1 = SkipConct(Multihead(LayerNorm(Qin;Kin;Vin)))
    M2 = SkipConct(Multihead(LayerNorm(M1;O;O)))
    L = SkipConct(FFN(LayerNorm(M2)))
    """

    def __init__(self, dim_model, total_res, heads_de, seq_len, dropout, device):
        super().__init__()
        self.seq_len = seq_len
        self.device = device
        self.embd_res = nn.Embedding(total_res + 1,
                                     embedding_dim=dim_model)  # response embedding, include a start token
        # self.embd_pos   = nn.Embedding(seq_len, embedding_dim = dim_model)                  #positional embedding
        self.multi_de1 = nn.MultiheadAttention(embed_dim=dim_model, num_heads=heads_de,
                                               dropout=dropout)  # M1 multihead for interaction embedding as q k v
        self.multi_de2 = nn.MultiheadAttention(embed_dim=dim_model, num_heads=heads_de,
                                               dropout=dropout)  # M2 multihead for M1 out, encoder out, encoder out as q k v
        self.ffn_en = transformer_FFN(dim_model, dropout)  # feed forward layer

        self.layer_norm1 = nn.LayerNorm(dim_model)
        self.layer_norm2 = nn.LayerNorm(dim_model)
        self.layer_norm3 = nn.LayerNorm(dim_model)

        self.dropout1 = Dropout(dropout)
        self.dropout2 = Dropout(dropout)
        self.dropout3 = Dropout(dropout)

    def forward(self, in_res, in_pos, en_out, first_block=True):

        ##
        if first_block:
            in_in = self.embd_res(in_res)

            # combining the embedings
            out = in_in + in_pos  # (b,n,d)
        else:
            out = in_res

        # in_pos = get_pos(self.seq_len)
        # in_pos = self.embd_pos(in_pos)

        out = out.permute(1, 0, 2)  # (n,b,d)# print('pre multi', out.shape)
        n, _, _ = out.shape

        # Multihead attention M1
        out = self.layer_norm1(out)
        skip_out = out
        out, attn_wt = self.multi_de1(out, out, out,
                                      attn_mask=ut_mask(self.device, seq_len=n))  # attention mask upper triangular
        out = self.dropout1(out)
        out = skip_out + out  # skip connection

        # Multihead attention M2
        en_out = en_out.permute(1, 0, 2)  # (b,n,d)-->(n,b,d)
        en_out = self.layer_norm2(en_out)
        skip_out = out
        out, attn_wt = self.multi_de2(out, en_out, en_out,
                                      attn_mask=ut_mask(self.device, seq_len=n))  # attention mask upper triangular
        out = self.dropout2(out)
        out = out + skip_out

        # feed forward
        out = out.permute(1, 0, 2)  # (b,n,d)
        out = self.layer_norm3(out)  # Layer norm
        skip_out = out
        out = self.ffn_en(out)
        out = self.dropout3(out)
        out = out + skip_out  # skip connection

        return out


from .base_mid2cache import BaseMid2Cache
import numpy as np
from itertools import chain
import torch


class M2C_GenQMat(BaseMid2Cache):
    def process(self, **kwargs):
        df_exer = kwargs['df_exer']
        cpt_count = len(set(list(chain(*df_exer['cpt_seq:token_seq'].to_list()))))
        # df_exer['cpt_multihot:token_seq'] = df_exer['cpt_seq:token_seq'].apply(
        #     lambda x: self.multi_hot(cpt_count, np.array(x)).tolist()
        # )
        kwargs['df_exer'] = df_exer
        tmp_df_exer = df_exer.set_index("exer_id:token")

        kwargs['Q_mat'] = torch.from_numpy(np.array(
            [self.multi_hot(cpt_count, tmp_df_exer.loc[exer_id]['cpt_seq:token_seq']).tolist()
              for exer_id in range(df_exer['exer_id:token'].max() + 1)]
        ))
        return kwargs

    @staticmethod
    def multi_hot(length, indices):
        multi_hot = np.zeros(length, dtype=np.int64)
        multi_hot[indices] = 1
        return multi_hot



r"""
FairCD
##########################################

Reference:
   Zheng Zhang et al. "Understanding and Improving Fairness in Cognitive Diagnosis" in SCIS 2023.


"""

from edustudio.model.CD import IRT, NCDM, MIRT
from edustudio.model.utils.components import MLP
import torch.nn as nn
import torch.nn.functional as F
import torch
from sklearn.preprocessing import LabelEncoder



class DiscriminatorForDiscreteSingleAttr(nn.Module):  
    def __init__(self, input_dim, output_dim, dnn_units, activation='relu', device='cuda:0'):  
        super().__init__()   
        self.mlp = MLP(  
        input_dim=input_dim,  
        output_dim=output_dim,  
        dnn_units=dnn_units,  
        activation=activation,  
        device=device  
        )  
        self.loss = nn.CrossEntropyLoss()  
        self.to(device)

    def forward(self, x, y):  
        """_summary_  
        Args:  
        x (_type_): BatchNum x FeatNum  
        y (_type_): (BatachNum,)  
        """  
        pd = self.mlp(x)  
        return self.loss(pd, y.flatten()) 


class FairCD_IRT(IRT):
    default_cfg = {
        'sensi_attr': 'gender:token'
    }

    def add_extra_data(self, **kwargs):
        super().add_extra_data(**kwargs)
        self.df_stu = kwargs['df_stu']
        self.df_stu_index= self.df_stu.set_index("stu_id:token")
        assert self.df_stu['stu_id:token'].max() + 1 == kwargs['dt_info']['stu_count']

        self.attr_name = self.modeltpl_cfg['sensi_attr']
        attr_sufix = self.attr_name.split(":")[-1]

        if attr_sufix == 'token':
            self.disc_cls = DiscriminatorForDiscreteSingleAttr
            self.pred_cls = DiscriminatorForDiscreteSingleAttr
            n_classes = self.df_stu[self.attr_name].nunique()
            self.out_dim = n_classes
            lbe = LabelEncoder()
            lbe.fit(self.df_stu[self.attr_name])
            self.label = torch.LongTensor(
                [lbe.transform([self.df_stu_index[self.attr_name].loc[sid]])[-1] for sid in range(kwargs['dt_info']['stu_count'])]
            ).to(self.device)
        else:
            raise NotImplementedError

    def build_model(self):
        super().build_model()
        self.bias_theta = nn.Embedding(self.n_user, 1)
        self.discriminator = self.disc_cls(
                    input_dim=1,
                    output_dim=self.out_dim,
                    dnn_units=[],
                    device=self.device
               )
        self.predictor = self.pred_cls(
                    input_dim=1,
                    output_dim=self.out_dim,
                    dnn_units=[],
                    device=self.device
               )
        self.logger.info(self)

    def get_g_parameters(self, recurse: bool = True):
        for name, param in self.named_parameters(recurse=recurse):
            if 'discriminator' not in name and 'predictor' not in name:
                yield param

    def get_d_parameters(self, recurse: bool = True):
        for name, param in self.named_parameters(recurse=recurse):
            if 'discriminator' in name or 'predictor' in name:
                yield param

    def get_adv_loss(self, **kwargs):
        stu_id = kwargs['stu_id'].unique()
        x = self.theta(stu_id)
        loss_adv_dict = {
            f'loss_dis_{self.attr_name}': self.discriminator(x, self.label[stu_id])
        }

        return loss_adv_dict

    def get_pre_loss(self, **kwargs):
        stu_id = kwargs['stu_id'].unique()
        x = self.bias_theta(stu_id)
        loss_adv_dict = {
            f'loss_dis_{self.attr_name}': self.predictor(x, self.label[stu_id])
        }

        return loss_adv_dict

    def get_loss_dict(self, **kwargs):
        loss_pre= self.get_pre_loss(**kwargs)
        loss_dis = self.get_adv_loss(**kwargs)

        return loss_pre, loss_dis

    def forward(self, stu_id, exer_id, **kwargs):
        theta = self.theta(stu_id)
        bias_theta = self.bias_theta(stu_id)
        theta = theta + bias_theta
        a = self.a(exer_id)
        b = self.b(exer_id)
        c = self.c if self.modeltpl_cfg['fix_c'] else self.c(exer_id).sigmoid()

        if self.modeltpl_cfg['diff_range'] is not None:
            b = self.modeltpl_cfg['diff_range'] * (torch.sigmoid(b) - 0.5)
        if self.modeltpl_cfg['a_range'] is not None:
            a = self.modeltpl_cfg['a_range'] * torch.sigmoid(a)
        else:
            a = F.softplus(a) # 让区分度大于0，保持单调性假设
        if torch.max(theta != theta) or torch.max(a != a) or torch.max(b != b):  # pragma: no cover
            raise ValueError('ValueError:theta,a,b may contains nan!  The diff_range or a_range is too large.')
        return self.irf(theta, a, b, c)

class FairCD_MIRT(MIRT):
    default_cfg = {
        'sensi_attr': 'gender:token'
    }

    def add_extra_data(self, **kwargs):
        super().add_extra_data(**kwargs)
        self.df_stu = kwargs['df_stu']
        self.df_stu_index= self.df_stu.set_index("stu_id:token")
        assert self.df_stu['stu_id:token'].max() + 1 == kwargs['dt_info']['stu_count']

        self.attr_name = self.modeltpl_cfg['sensi_attr']
        attr_sufix = self.attr_name.split(":")[-1]

        if attr_sufix == 'token':
            self.disc_cls = DiscriminatorForDiscreteSingleAttr
            self.pred_cls = DiscriminatorForDiscreteSingleAttr
            n_classes = self.df_stu[self.attr_name].nunique()
            self.out_dim = n_classes
            lbe = LabelEncoder()
            lbe.fit(self.df_stu[self.attr_name])
            self.label = torch.LongTensor(
                [lbe.transform([self.df_stu_index[self.attr_name].loc[sid]])[-1] for sid in range(kwargs['dt_info']['stu_count'])]
            ).to(self.device)
        else:
            raise NotImplementedError

    def build_model(self):
        super().build_model()
        self.bias_theta = nn.Embedding(self.n_user, self.emb_dim)
        self.discriminator = self.disc_cls(
                    input_dim=self.emb_dim,
                    output_dim=self.out_dim,
                    dnn_units=[int(self.emb_dim / 2 )],
                    device=self.device
               )
        self.predictor = self.pred_cls(
                    input_dim=self.emb_dim,
                    output_dim=self.out_dim,
                    dnn_units=[int(self.emb_dim / 2 )],
                    device=self.device
               )
        self.logger.info(self)

    def get_g_parameters(self, recurse: bool = True):
        for name, param in self.named_parameters(recurse=recurse):
            if 'discriminator' not in name and 'predictor' not in name:
                yield param

    def get_d_parameters(self, recurse: bool = True):
        for name, param in self.named_parameters(recurse=recurse):
            if 'discriminator' in name or 'predictor' in name:
                yield param

    def get_adv_loss(self, **kwargs):
        stu_id = kwargs['stu_id'].unique()
        x = self.theta(stu_id)
        loss_adv_dict = {
            f'loss_dis_{self.attr_name}': self.discriminator(x, self.label[stu_id])
        }

        return loss_adv_dict

    def get_pre_loss(self, **kwargs):
        stu_id = kwargs['stu_id'].unique()
        x = self.bias_theta(stu_id)
        loss_adv_dict = {
            f'loss_dis_{self.attr_name}': self.predictor(x, self.label[stu_id])
        }

        return loss_adv_dict

    def get_loss_dict(self, **kwargs):
        loss_pre= self.get_pre_loss(**kwargs)
        loss_dis = self.get_adv_loss(**kwargs)

        return loss_pre, loss_dis

    def forward(self, stu_id, exer_id, **kwargs):
        theta = self.theta(stu_id) + self.bias_theta(stu_id)
        a = self.a(exer_id)
        b = self.b(exer_id).flatten()

        if self.modeltpl_cfg['a_range'] is not None:
            a = self.modeltpl_cfg['a_range'] * torch.sigmoid(a)
        else:
            a = F.softplus(a) # 让区分度大于0，保持单调性假设
        if torch.max(theta != theta) or torch.max(a != a) or torch.max(b != b):  # pragma: no cover
            raise ValueError('ValueError:theta,a,b may contains nan!  The diff_range or a_range is too large.')
        return self.irf(theta, a, b)

class FairCD_NCDM(NCDM):
    default_cfg = {
        'sensi_attr': 'gender:token'
    }

    def add_extra_data(self, **kwargs):
        super().add_extra_data(**kwargs)
        self.df_stu = kwargs['df_stu']
        self.df_stu_index= self.df_stu.set_index("stu_id:token")
        assert self.df_stu['stu_id:token'].max() + 1 == kwargs['dt_info']['stu_count']

        self.attr_name = self.modeltpl_cfg['sensi_attr']
        attr_sufix = self.attr_name.split(":")[-1]

        if attr_sufix == 'token':
            self.disc_cls = DiscriminatorForDiscreteSingleAttr
            self.pred_cls = DiscriminatorForDiscreteSingleAttr
            n_classes = self.df_stu[self.attr_name].nunique()
            self.out_dim = n_classes
            lbe = LabelEncoder()
            lbe.fit(self.df_stu[self.attr_name])
            self.label = torch.LongTensor(
                [lbe.transform([self.df_stu_index[self.attr_name].loc[sid]])[-1] for sid in range(kwargs['dt_info']['stu_count'])]
            ).to(self.device)
        else:
            raise NotImplementedError

    def build_model(self):
        super().build_model()
        self.bias_student_emb = nn.Embedding(self.n_user, self.n_cpt)
        self.discriminator = self.disc_cls(
                    input_dim=self.n_cpt,
                    output_dim=self.out_dim,
                    dnn_units=[int(self.n_cpt / 2 )],
                    device=self.device
               )
        self.predictor = self.pred_cls(
                    input_dim=self.n_cpt,
                    output_dim=self.out_dim,
                    dnn_units=[int(self.n_cpt / 2 )],
                    device=self.device
               )
        self.logger.info(self)

    def get_g_parameters(self, recurse: bool = True):
        for name, param in self.named_parameters(recurse=recurse):
            if 'discriminator' not in name and 'predictor' not in name:
                yield param

    def get_d_parameters(self, recurse: bool = True):
        for name, param in self.named_parameters(recurse=recurse):
            if 'discriminator' in name or 'predictor' in name:
                yield param

    def get_adv_loss(self, **kwargs):
        stu_id = kwargs['stu_id'].unique()
        x = self.student_emb(stu_id)
        loss_adv_dict = {
            f'loss_dis_{self.attr_name}': self.discriminator(x, self.label[stu_id])
        }

        return loss_adv_dict

    def get_pre_loss(self, **kwargs):
        stu_id = kwargs['stu_id'].unique()
        x = self.bias_student_emb(stu_id)
        loss_adv_dict = {
            f'loss_dis_{self.attr_name}': self.predictor(x, self.label[stu_id])
        }

        return loss_adv_dict

    def get_loss_dict(self, **kwargs):
        loss_pre= self.get_pre_loss(**kwargs)
        loss_dis = self.get_adv_loss(**kwargs)

        return loss_pre, loss_dis

    def forward(self, stu_id, exer_id, **kwargs):
        # before prednet
        stu_emb = self.student_emb(stu_id) + self.bias_student_emb(stu_id)
        stat_emb = torch.sigmoid(stu_emb)
        k_difficulty = torch.sigmoid(self.k_difficulty(exer_id))
        e_difficulty = torch.sigmoid(self.e_difficulty(exer_id)) * self.modeltpl_cfg['disc_scale']
        # prednet
        input_knowledge_point = self.Q_mat[exer_id]
        input_x = e_difficulty * (stat_emb - k_difficulty) * input_knowledge_point
        pd = self.pd_net(input_x).sigmoid()
        return pd



from .gd_traintpl import GDTrainTPL
from edustudio.utils.common import UnifyConfig, set_same_seeds, tensor2npy
from edustudio.utils.callback import ModelCheckPoint, EarlyStopping, History, BaseLogger, Callback, CallbackList
from edustudio.model import BaseModel
import torch
from typing import Sequence
from collections import defaultdict
from tqdm import tqdm
import numpy as np
import shutil


class GeneralTrainTPL(GDTrainTPL):
    default_cfg = {
        'num_stop_rounds': 10,
        'early_stop_metrics': [('auc','max')],
        'best_epoch_metric': 'auc',
        'unsave_best_epoch_pth': True,
        'ignore_metrics_in_train': [],
        'batch_size': 32,
        'eval_batch_size': 32
    }

    def _check_params(self):
        super()._check_params()
        assert self.traintpl_cfg['best_epoch_metric'] in set(i[0] for i in self.traintpl_cfg['early_stop_metrics'])

    def one_fold_start(self, fold_id):
        super().one_fold_start(fold_id)
        # callbacks
        num_stop_rounds = self.traintpl_cfg['num_stop_rounds']
        es_metrics = self.traintpl_cfg["early_stop_metrics"]
        modelCheckPoint = ModelCheckPoint(
            es_metrics, save_folder_path=f"{self.frame_cfg.temp_folder_path}/pths/{fold_id}/"
        )
        es_cb = EarlyStopping(es_metrics, num_stop_rounds=num_stop_rounds, start_round=1)
        history_cb = History(folder_path=f"{self.frame_cfg.temp_folder_path}/history/{fold_id}", plot_curve=False)
        callbacks = [
            modelCheckPoint, es_cb, history_cb,
            BaseLogger(self.logger, group_by_contains=['loss'])
        ]
        self.callback_list = CallbackList(callbacks=callbacks, model=self.model, logger=self.logger)
        # evaltpls
        for evaltpl in self.evaltpls: 
            evaltpl.set_callback_list(self.callback_list)
            evaltpl.set_dataloaders(train_loader=self.train_loader, 
                                    valid_loader=self.valid_loader, 
                                    test_loader=self.test_loader
                                    )
        # train
        #set_same_seeds(self.traintpl_cfg['seed'])
        if self.valid_loader is not None:
            self.fit(train_loader=self.train_loader, valid_loader=self.valid_loader)
        else:
            self.logger.info("Without validation set, replace it with test set")
            self.fit(train_loader=self.train_loader, valid_loader=self.test_loader)
        
        metric_name = self.traintpl_cfg['best_epoch_metric']
        metric = [m for m in modelCheckPoint.metric_list if m.name == metric_name][0]
        if self.valid_loader is not None:
            # load best params
            fpth =  f"{self.frame_cfg.temp_folder_path}/pths/{fold_id}/best-epoch-{metric.best_epoch:03d}-for-{metric.name}.pth"
            self.model.load_state_dict(torch.load(fpth))

            metrics = self.inference(self.test_loader)
            for name in metrics: self.logger.info(f"{name}: {metrics[name]}")
            History.dump_json(metrics, f"{self.frame_cfg.temp_folder_path}/{fold_id}/result.json")
        else:
            # metrics = history_cb.log_as_time[metric.best_epoch]
            # load best params
            fpth =  f"{self.frame_cfg.temp_folder_path}/pths/{fold_id}/best-epoch-{metric.best_epoch:03d}-for-{metric.name}.pth"
            self.model.load_state_dict(torch.load(fpth))

            metrics = self.inference(self.test_loader)
            for name in metrics: self.logger.info(f"{name}: {metrics[name]}")
            History.dump_json(metrics, f"{self.frame_cfg.temp_folder_path}/{fold_id}/result.json")

        if self.traintpl_cfg['unsave_best_epoch_pth']: shutil.rmtree(f"{self.frame_cfg.temp_folder_path}/pths/")
        return metrics

    def fit(self, train_loader, valid_loader):
        self.model.train()
        optimizer = self.traintpl_cfg['optim']
        lr = self.traintpl_cfg['lr']
        weight_decay = self.traintpl_cfg['weight_decay']
        eps = self.traintpl_cfg['eps']
        self.optimizer = self._get_optim(optimizer=optimizer, lr=lr, weight_decay=weight_decay, eps=eps)
        self.callback_list.on_train_begin()
        for epoch in range(self.traintpl_cfg['epoch_num']):
            self.callback_list.on_epoch_begin(epoch + 1)
            logs = defaultdict(lambda: np.full((len(train_loader),), np.nan, dtype=np.float32))
            for batch_id, batch_dict in enumerate(
                    tqdm(train_loader, ncols=self.frame_cfg['TQDM_NCOLS'], desc="[EPOCH={:03d}]".format(epoch + 1))
            ):
                batch_dict = self.batch_dict2device(batch_dict)
                loss_dict = self.model.get_loss_dict(**batch_dict)
                loss = torch.hstack([i for i in loss_dict.values() if i is not None]).sum()
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
                for k in loss_dict: 
                    if loss_dict[k] is not None:
                        logs[k][batch_id] = loss_dict[k].item()

            for name in logs: logs[name] = float(np.nanmean(logs[name]))

            if valid_loader is not None:
                val_metrics = self.evaluate(valid_loader)
                logs.update({f"{metric}": val_metrics[metric] for metric in val_metrics})

            self.callback_list.on_epoch_end(epoch + 1, logs=logs)
            if self.model.share_callback_dict.get('stop_training', False):
                break
            
        self.callback_list.on_train_end()
    
    @torch.no_grad()
    def evaluate(self, loader):
        self.model.eval()
        stu_id_list = list(range(len(loader)))
        pd_list = list(range(len(loader)))
        gt_list = list(range(len(loader)))
        for idx, batch_dict in enumerate(tqdm(loader, ncols=self.frame_cfg['TQDM_NCOLS'], desc="[PREDICT]")):
            batch_dict = self.batch_dict2device(batch_dict)
            eval_dict = self.model.predict(**batch_dict)
            stu_id_list[idx] = batch_dict['stu_id']
            pd_list[idx] = eval_dict['y_pd']
            gt_list[idx] = eval_dict['y_gt'] if 'y_gt' in eval_dict else batch_dict['label']
        y_pd = torch.hstack(pd_list)
        y_gt = torch.hstack(gt_list)
        stu_id = torch.hstack(stu_id_list)

        eval_data_dict = {
            'stu_id': stu_id,
            'y_pd': y_pd,
            'y_gt': y_gt,
        }
        if hasattr(self.model, 'get_stu_status'):
            stu_stats_list = []
            idx = torch.arange(0, self.datatpl_cfg['dt_info']['stu_count']).to(self.traintpl_cfg['device'])
            for i in range(0,self.datatpl_cfg['dt_info']['stu_count'], self.traintpl_cfg['eval_batch_size']):
                batch_stu_id = idx[i:i+self.traintpl_cfg['eval_batch_size']]
                batch = self.model.get_stu_status(batch_stu_id)
                stu_stats_list.append(batch)
            stu_stats = torch.vstack(stu_stats_list)
            eval_data_dict.update({
                'stu_stats': tensor2npy(stu_stats),
            })
        if hasattr(self.datatpl, 'Q_mat'):
            eval_data_dict.update({
                'Q_mat': tensor2npy(self.datatpl.Q_mat)
            })
        eval_result = {}
        for evaltpl in self.evaltpls: eval_result.update(
                evaltpl.eval(ignore_metrics=self.traintpl_cfg['ignore_metrics_in_train'], **eval_data_dict)
            )
        return eval_result

    @torch.no_grad()
    def inference(self, loader):
        self.model.eval()
        stu_id_list = list(range(len(loader)))
        pd_list = list(range(len(loader)))
        gt_list = list(range(len(loader)))
        for idx, batch_dict in enumerate(tqdm(loader, ncols=self.frame_cfg['TQDM_NCOLS'], desc="[PREDICT]")):
            batch_dict = self.batch_dict2device(batch_dict)
            eval_dict = self.model.predict(**batch_dict)
            stu_id_list[idx] = batch_dict['stu_id']
            pd_list[idx] = eval_dict['y_pd']
            gt_list[idx] = eval_dict['y_gt'] if 'y_gt' in eval_dict else batch_dict['label']
        y_pd = torch.hstack(pd_list)
        y_gt = torch.hstack(gt_list)
        stu_id = torch.hstack(stu_id_list)

        eval_data_dict = {
            'stu_id': stu_id,
            'y_pd': y_pd,
            'y_gt': y_gt,
        }
        if hasattr(self.model, 'get_stu_status'):
            stu_stats_list = []
            idx = torch.arange(0, self.datatpl_cfg['dt_info']['stu_count']).to(self.traintpl_cfg['device'])
            for i in range(0,self.datatpl_cfg['dt_info']['stu_count'], self.traintpl_cfg['eval_batch_size']):
                batch_stu_id = idx[i:i+self.traintpl_cfg['eval_batch_size']]
                batch = self.model.get_stu_status(batch_stu_id)
                stu_stats_list.append(batch)
            stu_stats = torch.vstack(stu_stats_list)
            eval_data_dict.update({
                'stu_stats': tensor2npy(stu_stats),
            })
        if hasattr(self.datatpl, 'Q_mat'):
            eval_data_dict.update({
                'Q_mat': tensor2npy(self.datatpl.Q_mat)
            })
        eval_result = {}
        for evaltpl in self.evaltpls: eval_result.update(evaltpl.eval(**eval_data_dict))
        return eval_result



from .base_evaltpl import BaseEvalTPL
import torch
import pandas as pd
import numpy as np
from edustudio.utils.common import tensor2npy
from edustudio.utils.callback import ModeState
from tqdm import tqdm
import warnings

class InterpretabilityEvalTPL(BaseEvalTPL):
    """Student Cogntive Representation Interpretability Evaluation
    """
    default_cfg = {
        'use_metrics': ['doa_all'],
        'test_only_metrics': ['doa_all']
    }
    def __init__(self, cfg):
        super().__init__(cfg)
        self.name = self.__class__.__name__

    def eval(self, stu_stats:np.ndarray, Q_mat:np.ndarray, **kwargs):
        metric_result = {}
        ignore_metrics = kwargs.get('ignore_metrics', {})
        for metric_name in self.evaltpl_cfg[self.name]['use_metrics']:
            if metric_name not in ignore_metrics:
                if metric_name in self.evaltpl_cfg[self.name]['test_only_metrics'] and \
                    self.callback_list.mode_state != ModeState.END:
                   continue
               
                if metric_name == "doa_all":
                    metric_result[metric_name] = self._get_metrics(metric_name)(
                        self.stu_id_total,
                        self.exer_id_total,
                        self.label_id_total,
                        user_emb=stu_stats,
                        Q_mat=Q_mat
                    )
                elif metric_name == "doa_train_val":
                    metric_result[metric_name] = self._get_metrics(metric_name)(
                        self.stu_id_train_val,
                        self.exer_id_train_val,
                        self.label_id_train_val,
                        user_emb=stu_stats,
                        Q_mat=Q_mat
                    )
                elif metric_name == "doa_test":
                    metric_result[metric_name] = self._get_metrics(metric_name)(
                        self.test_loader.dataset.dict_main['stu_id'],
                        self.test_loader.dataset.dict_main['exer_id'],
                        self.test_loader.dataset.dict_main['label'],
                        user_emb=stu_stats,
                        Q_mat=Q_mat
                    )
                elif metric_name == "doc_all":
                    metric_result[metric_name] = self._get_metrics(metric_name)(
                        stu_stats,
                        self._gen_hit_mat(
                            self.stu_id_total,
                            self.exer_id_total,
                            Q_mat                        
                        ),
                        self.log_mat_total,
                        Q_mat
                    )
                elif metric_name == "doc_train_val":
                    metric_result[metric_name] = self._get_metrics(metric_name)(
                        stu_stats,
                        self._gen_hit_mat(
                            self.stu_id_train_val,
                            self.exer_id_train_val,
                            Q_mat 
                        ),
                        self.log_mat_train_val,
                        Q_mat
                    )
                elif metric_name == "doc_test":
                    metric_result[metric_name] = self._get_metrics(metric_name)(
                        stu_stats,
                        self._gen_hit_mat(
                            self.test_loader.dataset.dict_main['stu_id'],
                            self.test_loader.dataset.dict_main['exer_id'],
                            Q_mat
                        ),
                        self.log_mat_test,
                        Q_mat
                    )
                else:
                    raise ValueError(f"unknown metric_name: {metric_name}")
        return metric_result

    def _get_metrics(self, metric):
        if metric == "doa_all":
            return self.doa_report
        elif metric == "doa_test":
            return self.doa_report
        elif metric == "doa_train_val":
            return self.doa_report
        elif metric == "doc_all":
            return self.degree_of_consistency
        elif metric == "doc_test":
            return self.degree_of_consistency
        elif metric == "doc_train_val":
            return self.degree_of_consistency
        else:
            raise NotImplementedError
        
    def set_dataloaders(self, train_loader, test_loader, valid_loader=None):
        super().set_dataloaders(train_loader, test_loader, valid_loader)

        if self.valid_loader:
            self.stu_id_total = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['stu_id'],
                self.valid_loader.dataset.dict_main['stu_id'],
                self.test_loader.dataset.dict_main['stu_id'],
            ]))
            self.exer_id_total = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['exer_id'],
                self.valid_loader.dataset.dict_main['exer_id'],
                self.test_loader.dataset.dict_main['exer_id'],
            ]))

            self.label_id_total = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['label'],
                self.valid_loader.dataset.dict_main['label'],
                self.test_loader.dataset.dict_main['label'],
            ]))

            self.stu_id_train_val = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['stu_id'],
                self.valid_loader.dataset.dict_main['stu_id'],
            ]))
            self.exer_id_train_val = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['exer_id'],
                self.valid_loader.dataset.dict_main['exer_id'],
            ]))

            self.label_id_train_val = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['label'],
                self.valid_loader.dataset.dict_main['label'],
            ]))
        else:
            self.stu_id_total = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['stu_id'],
                self.test_loader.dataset.dict_main['stu_id'],
            ]))
            self.exer_id_total = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['exer_id'],
                self.test_loader.dataset.dict_main['exer_id'],
            ]))

            self.label_id_total = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['label'],
                self.test_loader.dataset.dict_main['label'],
            ]))

            self.stu_id_train_val = tensor2npy(self.train_loader.dataset.dict_main['stu_id'])
            self.exer_id_train_val = tensor2npy(self.train_loader.dataset.dict_main['exer_id'])
            self.label_id_train_val = tensor2npy(self.train_loader.dataset.dict_main['label'])

        self.log_mat_total = self._gen_log_mat(self.stu_id_total, self.exer_id_total, self.label_id_total)
        self.log_mat_train_val = self._gen_log_mat(self.stu_id_train_val, self.exer_id_train_val, self.label_id_train_val)
        self.log_mat_test = self._gen_log_mat(
            tensor2npy(self.test_loader.dataset.dict_main['stu_id']),
            tensor2npy(self.test_loader.dataset.dict_main['exer_id']),
            tensor2npy(self.test_loader.dataset.dict_main['label']),
        )

    def doa_report(self, stu_id, exer_id, label, user_emb, Q_mat):
        knowledges = []
        knowledge_item = []
        knowledge_user = []
        knowledge_truth = []
        knowledge_theta = []
        for user, item, score in zip(stu_id, exer_id, label):
            theta = user_emb[user]
            knowledge = Q_mat[item]
            if isinstance(theta, list) or isinstance(theta, np.ndarray):
                for i, (theta_i, knowledge_i) in enumerate(zip(theta, knowledge)):
                    if knowledge_i == 1: 
                        knowledges.append(i) # 知识点ID
                        knowledge_item.append(item) # Item ID
                        knowledge_user.append(user) # User ID
                        knowledge_truth.append(score) # score
                        knowledge_theta.append(theta_i) # matser
            else:  # pragma: no cover
                for i, knowledge_i in enumerate(knowledge):
                    if knowledge_i == 1:
                        knowledges.append(i)
                        knowledge_item.append(item)
                        knowledge_user.append(user)
                        knowledge_truth.append(score)
                        knowledge_theta.append(theta)

        knowledge_df = pd.DataFrame({
            "knowledge": knowledges,
            "user_id": knowledge_user,
            "item_id": knowledge_item,
            "score": knowledge_truth,
            "theta": knowledge_theta
        })
        knowledge_ground_truth = []
        knowledge_prediction = []
        for _, group_df in knowledge_df.groupby("knowledge"):
            _knowledge_ground_truth = []
            _knowledge_prediction = []
            for _, item_group_df in group_df.groupby("item_id"):
                _knowledge_ground_truth.append(item_group_df["score"].values)
                _knowledge_prediction.append(item_group_df["theta"].values)
            knowledge_ground_truth.append(_knowledge_ground_truth)
            knowledge_prediction.append(_knowledge_prediction)

        return self.doa_eval(knowledge_ground_truth, knowledge_prediction)


    def doa_eval(self, y_true, y_pred):
        doa = []
        doa_support = 0
        z_support = 0
        for knowledge_label, knowledge_pred in zip(y_true, y_pred):
            _doa = 0
            _z = 0
            for label, pred in zip(knowledge_label, knowledge_pred):
                if sum(label) == len(label) or sum(label) == 0:
                    continue
                pos_idx = []
                neg_idx = []
                for i, _label in enumerate(label): # 找出所有(1, 0) pair
                    if _label == 1:
                        pos_idx.append(i)
                    else:
                        neg_idx.append(i)
                pos_pred = pred[pos_idx]
                neg_pred = pred[neg_idx]
                invalid = 0
                for _pos_pred in pos_pred:
                    _doa += len(neg_pred[neg_pred < _pos_pred])
                    invalid += len(neg_pred[neg_pred == _pos_pred])
                _z += (len(pos_pred) * len(neg_pred)) - invalid
            if _z > 0:
                doa.append(_doa / _z)
                z_support += _z # 有效pair个数
                doa_support += 1 # 有效doa
        return float(np.mean(doa))


    def _gen_log_mat(self, uid, iid, label):
        n_stu = self.datatpl_cfg['dt_info']['stu_count']
        n_exer = self.datatpl_cfg['dt_info']['exer_count']
        label = label.copy()
        label[label == 0] = -1
        log_mat = np.zeros((n_stu, n_exer), dtype=np.float32)
        log_mat[uid, iid] = label
        return log_mat


    def _gen_hit_mat(self, uid, iid, Q_mat):
        n_stu = self.datatpl_cfg['dt_info']['stu_count']
        n_exer = self.datatpl_cfg['dt_info']['exer_count']
        n_cpt = self.datatpl_cfg['dt_info']['cpt_count']

        tmp_df = pd.DataFrame({
            'uid': uid,
            'iid': iid
        })

        # assert tmp_df['uid'].nunique() == n_stu

        hit_df = tmp_df.groupby('uid').agg(list).apply(lambda x: list(Q_mat[np.array(x['iid'])].sum(axis=0)), axis=1)
        if tmp_df['uid'].nunique() != n_stu:
            zeros = list(np.zeros((n_cpt, )))
            new_index = list(range(n_stu))
            hit_df = hit_df.reindex(new_index, fill_value=zeros)
        hit_df = hit_df.sort_index()

        return np.array(hit_df.to_list())

    """
        reference code: https://github.com/CSLiJT/ID-CDF/blob/main/tools.py
    """

    def degree_of_consistency_official(self, theta_mat: np.array, user_know_hit: np.array, \
                              log_mat: np.array, Q_mat: np.array, know_list=None):
        '''
        theta_mat: (n_user, n_know): the diagnostic result matrix
        user_know_hit: (n_user, n_know): the (i,j) element indicate \
            the number of hits of the i-th user on the j-th attribute
        log_mat: (n_user, n_exer): the matrix indicating whether the \
            student has correctly answered the exercise (+1) or not(-1)
        Q_mat: (n_exer, n_know)
        '''
        n_user, n_know = theta_mat.shape
        n_exer = log_mat.shape[1]
        doa_all = []
        know_list = list(range(n_know)) if know_list is None else know_list
        for know_id in tqdm(know_list, desc='compute_DOC', ncols=100):
            Z = 1e-9
            dm = 0
            exer_list = np.where(Q_mat[:, know_id] > 0)[0]
            user_list = np.where(user_know_hit[:, know_id] > 0)[0]
            n_u_k = len(user_list)
            # pbar = tqdm(total = n_u_k * (n_u_k - 1), desc='know_id = %d'%know_id)
            for a in user_list:
                for b in user_list:
                    # if m_ak != m_bk, then either m_ak > m_bk or m_bk > m_ak
                    if a == b:
                        continue
                    Z += (theta_mat[a, know_id] > theta_mat[b, know_id])
                    nab = 1e-9
                    dab = 1e-9
                    for exer_id in exer_list:
                        Jab = (log_mat[a, exer_id] * log_mat[b, exer_id] != 0)
                        nab += Jab * (log_mat[a, exer_id] > log_mat[b, exer_id])
                        dab += Jab * (log_mat[a, exer_id] != log_mat[b, exer_id])
                    dm += (theta_mat[a, know_id] > theta_mat[b, know_id]) * nab / dab
                    # pbar.update(1)

            doa = dm / Z
            doa_all.append(doa)
        return np.mean(doa_all)


    def degree_of_consistency(self, theta_mat: np.array, user_know_hit: np.array, \
                              log_mat: np.array, Q_mat: np.array, know_list=None):
        '''
        theta_mat: (n_user, n_know): the diagnostic result matrix
        user_know_hit: (n_user, n_know): the (i,j) element indicate \
            the number of hits of the i-th user on the j-th attribute
        log_mat: (n_user, n_exer): the matrix indicating whether the \
            student has correctly answered the exercise (+1) or not(-1)
        Q_mat: (n_exer, n_know)
        '''
        n_user, n_know = theta_mat.shape
        doa_all = []
        know_list = list(range(n_know)) if know_list is None else know_list
        for know_id in tqdm(know_list, desc='compute_DOC', ncols=100):
            Z = 0
            dm = 0
            exer_list = np.where(Q_mat[:, know_id] > 0)[0]
            user_list = np.where(user_know_hit[:, know_id] > 0)[0]
            n_user_hit = len(user_list)
            Matrics_NAB = np.zeros((n_user_hit, 1))
            Matrics_DAB = np.zeros((n_user_hit, 1))
            Matrics_Z = np.zeros((n_user_hit, 1))
            exerShape = len(exer_list)

            ###/////////###
            Simply_log_mat = np.zeros((n_user_hit, exerShape))
            Column_Z = np.zeros((n_user_hit))
            for a in range(n_user_hit):
                for b in range(exerShape):
                    Simply_log_mat[a, b] = log_mat[user_list[a], exer_list[b]]
                Column_Z[a] = theta_mat[user_list[a], know_id]
            ###/////////###

            for a in range(n_user_hit):
                Matrics_NAB_column = np.zeros((n_user_hit, 1))
                Matrics_DAB_column = np.zeros((n_user_hit, 1))
                Matrics_Z_column = np.zeros((n_user_hit, 1))
                for exer_matrics_num in range(exerShape):
                    Colum_exer_matrics = Simply_log_mat[:, exer_matrics_num]
                    Matrics_JAB_column = (Colum_exer_matrics * Simply_log_mat[a, exer_matrics_num] != 0).astype(
                        int).reshape(-1, 1)
                    if Colum_exer_matrics[a] == 1:
                        Matrics_NAB_column += (Colum_exer_matrics == -1).astype(int).reshape(-1,
                                                                                             1) * Matrics_JAB_column
                    Matrics_DAB_column += (Colum_exer_matrics != Simply_log_mat[a, exer_matrics_num]).astype(
                        int).reshape(-1, 1) * Matrics_JAB_column

                Matrics_NAB = np.concatenate([Matrics_NAB, Matrics_NAB_column], axis=1)
                Matrics_DAB = np.concatenate([Matrics_DAB, Matrics_DAB_column], axis=1)

                Matrics_Z_column = (Column_Z[a] - Column_Z).reshape(-1, 1)
                Matrics_Z = np.concatenate([Matrics_Z, Matrics_Z_column], axis=1)

            Matrics_Z = (Matrics_Z > 0).astype(int)[:, 1:n_user + 1]
            Matrics_NAB = Matrics_NAB[:, 1:n_user + 1]
            Matrics_DAB = Matrics_DAB[:, 1:n_user + 1]

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                Matrics_DM_divided = Matrics_NAB / Matrics_DAB

            Matrics_DM_divided = np.nan_to_num(Matrics_DM_divided, nan=1)
            Matrics_DM = Matrics_Z * Matrics_DM_divided
            Z = np.sum(Matrics_Z, dtype=np.float64)
            dm = round(np.sum(Matrics_DM, dtype=np.float64), 20)

            doa = round(dm / (Z+1e-9), 20)
            doa_all.append(doa)
        return np.mean(doa_all)



from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F


class SAKT(GDBaseModel):
    default_cfg = {
        'emb_size': 128,
        'n_attn_heads': 8,
        'max_length': 100,
        'dropout_rate': 0.2,
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']

    def build_model(self):
        self.M = nn.Embedding(self.n_item * 2, self.modeltpl_cfg['emb_size'])
        self.E = nn.Embedding(self.n_item, self.modeltpl_cfg['emb_size'])
        self.P = nn.Parameter(torch.Tensor(self.modeltpl_cfg['max_length'] - 1, self.modeltpl_cfg['emb_size']))

        self.attn = nn.MultiheadAttention(
            self.modeltpl_cfg['emb_size'], self.modeltpl_cfg['n_attn_heads'], dropout=self.modeltpl_cfg['dropout_rate']
        )
        self.attn_dropout = nn.Dropout(self.modeltpl_cfg['dropout_rate'])
        self.attn_layer_norm = nn.LayerNorm(self.modeltpl_cfg['emb_size'])

        self.FFN = nn.Sequential(
            nn.Linear(self.modeltpl_cfg['emb_size'], self.modeltpl_cfg['emb_size']),
            nn.ReLU(),
            nn.Dropout(self.modeltpl_cfg['dropout_rate']),
            nn.Linear(self.modeltpl_cfg['emb_size'], self.modeltpl_cfg['emb_size']),
            nn.Dropout(self.modeltpl_cfg['dropout_rate']),
        )
        self.FFN_layer_norm = nn.LayerNorm(self.modeltpl_cfg['emb_size'])

        self.pred = nn.Linear(self.modeltpl_cfg['emb_size'], 1)

    def forward(self, exer_seq, label_seq, **kwargs):
        q, r, qry = exer_seq[:, :-1].to(torch.int64), label_seq[:, :-1].to(torch.int64), exer_seq[:, 1:].to(torch.int64)

        x = q + self.n_item* r

        M = self.M(x).permute(1, 0, 2)
        E = self.E(qry).permute(1, 0, 2)
        P = self.P.unsqueeze(1)
        M = M + P

        causal_mask = torch.triu(
            torch.ones([E.shape[0], M.shape[0]]), diagonal=1
        ).bool().to(self.device)

        S, attn_weights = self.attn(E, M, M, attn_mask=causal_mask)
        

        S = self.attn_dropout(S)
        S = S.permute(1, 0, 2)
        M = M.permute(1, 0, 2)
        E = E.permute(1, 0, 2)

        S = self.attn_layer_norm(S + M + E)

        F = self.FFN(S)
        F = self.FFN_layer_norm(F + S)

        p = self.pred(F).squeeze(-1)

        return p, attn_weights
    
    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd, _ = self(**kwargs)
        # y_pd = y_pd[:, :-1].gather(
        #     index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=2
        # ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': torch.sigmoid(y_pd),
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd, _ = self(**kwargs)  # use forward
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        criterion = nn.BCEWithLogitsLoss()
        loss = criterion(
            input=y_pd, target=y_gt
        )
        return {
            'loss_main': loss
        }
        

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)


        